#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import sys
import logging
import seaborn as sns
from crispy.Utils import Utils
from crispy.QCPlot import QCplot
from crispy.CrispyPlot import CrispyPlot
from crispy.Enrichment import SSGSEA, GSEAplot
from crispy.CopyNumberCorrection import Crispy, CrispyGaussian


__version__ = "0.5.3"

# - SET STYLE
sns.set(
    style="ticks",
    context="paper",
    font_scale=0.75,
    font="sans-serif",
    rc=CrispyPlot.SNS_RC,
)

# - Logging
__name__ = "Crispy"

logger = logging.getLogger(__name__)

if not logger.handlers:
    ch = logging.StreamHandler(sys.stdout)
    ch.setFormatter(logging.Formatter("[%(asctime)s - %(levelname)s]: %(message)s"))
    logger.addHandler(ch)
    logger.setLevel(logging.INFO)
    logger.propagate = False

# - HANDLES
__all__ = [
    "Crispy",
    "CrispyGaussian",
    "QCplot",
    "CrispyPlot",
    "SSGSEA",
    "GSEAplot",
    "Utils",
    "__version__",
    "logger",
]



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import numpy as np
import pandas as pd
import crispy as cy
import matplotlib.pyplot as plt
from pybedtools import BedTool
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import WhiteKernel, ConstantKernel, RBF


PSEUDO_COUNT = 0.5

LOW_COUNT_THRES = 30

COPY_NUMBER_COLUMNS = ["Chr", "Start", "End", "copy_number"]

CRISPR_LIB_COLUMNS = ["Chr", "Start", "End"]

BED_COLUMNS = [
    "Chr",
    "Start",
    "End",
    "copy_number",
    "sgRNA_Chr",
    "sgRNA_Start",
    "sgRNA_End",
    "fold_change",
    "sgRNA_ID",
]


class Crispy:
    def __init__(self, sgrna_fc, library, copy_number, exclude_heterochromosomes=False):
        f"""
        Initialise a Crispy processing pipeline object

        :param copy_number: pandas.DataFrame
            Copy-number segments, must have these columns {COPY_NUMBER_COLUMNS}

        :param library: pandas.DataFrame
            CRISPR library, must have these columns {CRISPR_LIB_COLUMNS}

        """
        self.library = library
        self.library.index.name = "index"

        self.sgrna_fc = sgrna_fc
        self.copy_number = copy_number
        self.gpr = None
        self.exclude_heterochromosomes = exclude_heterochromosomes

    def correct(
        self,
        x_features=None,
        y_feature="fold_change",
        n_sgrna=10,
    ):
        """
        Main pipeline function to process data from raw counts to corrected fold-changes.

        :param x_features: list
            Columns to be used as feature(s)

        :param y_feature: str
            Dependent variable

        :param n_sgrna: int
            Minimum number of guides per segment

        :param round_dec: int
            Number of decimal places for floating numbers. If equals to None no rounding is performed

        :return: pandas.DataFrame
        """
        assert self.copy_number is not None, "Copy-number information not provided"

        if x_features is None:
            x_features = ["ratio"]

        elif type(x_features) == str:
            x_features = [x_features]

        else:
            x_features = x_features

        # - Intersect sgRNAs genomic localisation with copy-number segments
        bed_df = self.intersect_sgrna_copynumber()
        bed_df = bed_df.dropna(subset=x_features + [y_feature])

        # - Fit Gaussian Process on segment fold-changes
        self.gpr = CrispyGaussian(bed_df, n_sgrna=n_sgrna)
        self.gpr = self.gpr.fit(x=x_features, y=y_feature)

        bed_df["gp_mean"] = self.gpr.predict(bed_df[x_features])

        # - Correct fold-change by subtracting the estimated mean
        bed_df["corrected"] = bed_df.eval("fold_change - gp_mean")

        # - Add gene
        bed_df["gene"] = (
            self.library.reindex(bed_df["sgRNA_ID"])["Approved_Symbol"].values
        )

        return bed_df

    def calculate_ploidy(self, df):
        """
        Estimate ploidy and chromosomes number of copies from copy-number segments. Mean copy-number is
        weigthed by the length of the segment.

        :param df: pandas.DataFrame or pybedtools.BedTool

        :return: (pandas.Series, float)
            Chromosome copies, ploidy
        """
        if type(df) == BedTool:
            df = df.to_dataframe().rename(
                columns={"name": "copy_number", "chrom": "chr"}
            )

        if self.exclude_heterochromosomes:
            df = df[~df["Chr"].isin(["chrX", "chrY"])]

        df = df.assign(length=df["End"] - df["Start"])
        df = df.assign(cn_by_length=df["length"] * (df["copy_number"] + 1))

        chrm = (
            df.groupby("Chr")["cn_by_length"]
            .sum()
            .divide(df.groupby("Chr")["length"].sum())
            - 1
        )
        chrm.index = [str(i) if type(i) is not str else i for i in chrm.index]

        ploidy = (df["cn_by_length"].sum() / df["length"].sum()) - 1

        return chrm, ploidy

    def get_df_copy_number(self):
        """
        Create BedTool object from copy-number segments data-frame

        :return: pybedtools.BedTool
        """
        return self.copy_number[COPY_NUMBER_COLUMNS]

    def get_df_library(self):
        """
        Create BedTool object for sgRNAs in CRISPR-Cas9 library

        :param fold_change: pandas.Series, optional
            Estimated fold-changes

        :return: pybedtools.BedTool
        """
        lib = self.library[CRISPR_LIB_COLUMNS].dropna()

        lib = lib[lib.index.isin(self.sgrna_fc.index)]
        lib = lib.assign(fold_change=self.sgrna_fc[lib.index].values)

        lib = lib.assign(Start=lib["Start"].astype(int).values)
        lib = lib.assign(End=lib["End"].astype(int).values)

        return lib.reset_index()[CRISPR_LIB_COLUMNS + ["fold_change", "index"]]

    def intersect_sgrna_copynumber(self):
        """
        Intersect sgRNAs fold-changes with copy-number segments

        :param fold_change: pandas.Series
            sgRNAs fold-changes

        :return: pandas.DataFrame
            DataFrame containing the intersection of the copy-number segments with the sgRNAs
            of the CRISPR library

        """
        # Build copy-number and sgRNAs data frames
        df_cn = self.get_df_copy_number()
        df_sg = self.get_df_library()

        # Build beds
        bed_cn = BedTool(df_cn.to_string(index=False, header=False), from_string=True).sort()
        bed_sg = BedTool(df_sg.to_string(index=False, header=False), from_string=True).sort()

        # Intersect copy-number segments with sgRNAs
        bed_df = bed_cn.intersect(bed_sg, wa=True, wb=True).to_dataframe(names=BED_COLUMNS)
        bed_df["sgRNA_ID"] = [f"{i:.0f}".split(".")[0] if type(i) is not str else i for i in bed_df["sgRNA_ID"]]

        # Calculate chromosome copies and cell ploidy
        chrm, ploidy = self.calculate_ploidy(df_cn)

        bed_df = bed_df.assign(chr_copy=chrm[bed_df["Chr"].astype(str)].values)
        bed_df = bed_df.assign(ploidy=ploidy)

        # Calculate copy-number ratio
        bed_df = bed_df.assign(ratio=bed_df.eval("copy_number / chr_copy"))

        # Calculate segment length
        bed_df = bed_df.assign(len=bed_df.eval("End - Start"))
        bed_df = bed_df.assign(len_log2=bed_df["len"].apply(np.log2))

        return bed_df


class CrispyGaussian(GaussianProcessRegressor):
    SEGMENT_COLUMNS = ["Chr", "Start", "End"]

    SEGMENT_AGG_FUN = dict(
        fold_change=np.mean,
        copy_number=np.mean,
        ratio=np.mean,
        chr_copy=np.mean,
        ploidy=np.mean,
        len=np.mean,
        len_log2=np.mean,
        sgRNA_ID="count",
    )

    def __init__(
        self,
        bed_df,
        kernel=None,
        n_sgrna=10,
        alpha=1e-10,
        n_restarts_optimizer=3,
        optimizer="fmin_l_bfgs_b",
        normalize_y=False,
        copy_x_train=False,
        random_state=None,
    ):

        self.bed_seg = bed_df.groupby(self.SEGMENT_COLUMNS).agg(self.SEGMENT_AGG_FUN)
        self.n_sgrna = n_sgrna

        super().__init__(
            alpha=alpha,
            kernel=self.get_default_kernel() if kernel is None else kernel,
            optimizer=optimizer,
            normalize_y=normalize_y,
            random_state=random_state,
            copy_X_train=copy_x_train,
            n_restarts_optimizer=n_restarts_optimizer,
        )

    @property
    def _constructor(self):
        return CrispyGaussian

    @staticmethod
    def get_default_kernel():
        """
        Deafult Gaussian Process kernel

        :return: sklearn.gaussian_process.kernels.Sum
        """
        return ConstantKernel() * RBF() + WhiteKernel()

    def fit(self, x, y, train_idx=None):
        if train_idx is not None:
            x = self.bed_seg.iloc[train_idx].query(f"sgRNA_ID >= {self.n_sgrna}")[x]
            y = self.bed_seg.iloc[train_idx].query(f"sgRNA_ID >= {self.n_sgrna}")[y]

        else:
            x = self.bed_seg.query(f"sgRNA_ID >= {self.n_sgrna}")[x]
            y = self.bed_seg.query(f"sgRNA_ID >= {self.n_sgrna}")[y]

        return super().fit(x, y)

    def predict(self, x=None, return_std=False, return_cov=False):
        if x is None:
            x = self.bed_seg[["ratio"]]

        return super().predict(x, return_std=return_std, return_cov=return_cov)

    def score(self, x=None, y=None, sample_weight=None):
        if x is None:
            x = self.bed_seg[["ratio"]]
        elif type(x) == list:
            x = self.bed_seg[x]

        if y is None:
            y = self.bed_seg["fold_change"]

        return super().score(x, y, sample_weight=sample_weight)

    def plot(self, x_feature="ratio", y_feature="fold_change", ax=None):
        """
        Plot original data and GPR

        :param x_feature: str, x axis variable
        :param y_feature: str, y axis variable
        :param ax: matplotlib.plot
        :return: matplotlib.plot
        """

        if ax is None:
            ax = plt.gca()

        # - Data
        x, y = (
            self.bed_seg.query(f"sgRNA_ID >= {self.n_sgrna}")[x_feature],
            self.bed_seg.query(f"sgRNA_ID >= {self.n_sgrna}")[y_feature],
        )
        x_, y_ = (
            self.bed_seg.query(f"sgRNA_ID < {self.n_sgrna}")[x_feature],
            self.bed_seg.query(f"sgRNA_ID < {self.n_sgrna}")[y_feature],
        )

        x_pred = np.arange(0, x.max(), 0.1)
        y_pred, y_pred_std = self.predict(x_pred.reshape(-1, 1), return_std=True)

        # - Plot
        # Segments used for fitting
        ax.scatter(
            x,
            y,
            c=cy.QCplot.PAL_DBGD[0],
            alpha=0.7,
            edgecolors="white",
            lw=0.3,
            label=f"#(sgRNA_ID) >= {self.n_sgrna}",
        )

        # Segments not used for fitting
        plt.scatter(
            x_,
            y_,
            c=cy.QCplot.PAL_DBGD[0],
            marker="X",
            alpha=0.3,
            edgecolors="white",
            lw=0.3,
            label=f"#(sgRNA_ID) < {self.n_sgrna}",
        )

        # Plot GP fit
        # GP fit
        plt.plot(
            x_pred, y_pred, ls="-", lw=1.0, c=cy.QCplot.PAL_DBGD[1], label="GPR mean"
        )
        plt.fill_between(
            x_pred,
            y_pred - y_pred_std,
            y_pred + y_pred_std,
            alpha=0.2,
            color=cy.QCplot.PAL_DBGD[1],
            lw=0,
        )

        # Misc
        plt.axhline(0, ls=":", color=cy.QCplot.PAL_DBGD[2], lw=0.3, zorder=0)

        plt.xlabel(f"Segment\n{x_feature}")
        plt.ylabel(f"Segment\nmean {y_feature}")

        plt.title(f"{self.kernel_}", fontsize=6)

        plt.legend(frameon=False)

        return ax



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import os
import h5py
import gseapy
import logging
import matplotlib
import numpy as np
import pandas as pd
import pkg_resources
import seaborn as sns
import matplotlib.pyplot as plt
from adjustText import adjust_text
from crispy.Enrichment import Enrichment
from crispy.CrispyPlot import CrispyPlot
from mofapy2.run.entry_point import entry_point
from sklearn.linear_model import LinearRegression

LOG = logging.getLogger("Crispy")
DPATH = pkg_resources.resource_filename("crispy", "data")


class MOFA:
    def __init__(
        self,
        views=None,
        groupby=None,
        likelihoods=None,
        factors_n=10,
        covariates=None,
        fit_intercept=True,
        scale_views=True,
        scale_groups=True,
        iterations=1000,
        convergence_mode="slow",
        use_overlap=True,
        startELBO=1,
        freqELBO=1,
        dropR2=None,
        verbose=1,
        from_file=None,
    ):
        """
        This is a wrapper of MOFA to perform multi-omics integrations of the GDSC data-sets. 
        
            - Multiple groups are NOT supported
            - Only samples part of all views are considered
        
        :param views: dict(str: pandas.DataFrame)
        """

        self.verbose = verbose
        self.from_file = from_file

        self.factors_n = factors_n
        self.factors_labels = [f"F{i + 1}" for i in range(factors_n)]

        self.likelihoods = likelihoods

        self.views = views
        self.scale_views = scale_views
        self.scale_groups = scale_groups
        self.views_labels = list(self.views)

        self.use_overlap = use_overlap

        self.iterations = iterations
        self.convergence_mode = convergence_mode
        self.startELBO = startELBO
        self.freqELBO = freqELBO
        self.dropR2 = dropR2

        # Covariates
        self.covariates = covariates
        self.fit_intercept = fit_intercept

        # Samples
        self.samples = set.intersection(
            *[set(self.views[v]) for v in self.views_labels]
        )

        if self.covariates is not None:
            LOG.info(f"Covariates provided N={len(self.covariates)}")
            for k, v in self.covariates.items():
                self.samples = self.samples.intersection(set(v.index))

        self.samples = list(self.samples)
        LOG.info(f"Overlaping samples: {len(self.samples)}")

        # Reduce to overlaping samples
        if self.use_overlap:
            for k, df in self.views.items():
                self.views[k] = df[self.samples]

        # Info
        for k, df in self.views.items():
            LOG.info(f"View {k}: {df.shape}")

        # Regress-out covariates
        if self.covariates is not None:
            self.views = self.regress_out_covariates(fit_intercept=self.fit_intercept)

        # RUN MOFA
        # Prepare data, melt & tidy
        self.data = []

        for k in self.views_labels:
            df = self.views[k].copy()
            df.index.name = "feature"
            df.columns.name = "sample"

            df = df.unstack().rename("value").reset_index()

            if groupby is not None:
                df["group"] = groupby.reindex(df["sample"]).values

            else:
                df = df.assign(group="gdsc")

            self.data.append(df.assign(view=k))

        self.data = pd.concat(self.data, ignore_index=True)
        self.data = self.data[["sample", "group", "feature", "value", "view"]].dropna()

        # Initialise entry point
        self.ep = entry_point()

        # Set data options
        self.ep.set_data_options(
            scale_groups=self.scale_groups, scale_views=self.scale_views
        )
        self.ep.set_data_df(self.data, likelihoods=self.likelihoods)

        # Set model options
        self.ep.set_model_options(factors=self.factors_n)

        # Set training options
        self.ep.set_train_options(
            iter=self.iterations,
            convergence_mode=self.convergence_mode,
            startELBO=self.startELBO,
            freqELBO=self.freqELBO,
            dropR2=self.dropR2,
            verbose=verbose > 0,
        )

        # Run MOFA
        self.ep.build()

        if not os.path.isfile(self.from_file):
            self.ep.run()
            self.save_hdf5(self.from_file)

        self.mofa_file = h5py.File(self.from_file, "r")

        self.factors = self.get_factors(self.mofa_file)
        self.weights = self.get_weights(self.mofa_file)
        self.rsquare = self.get_rsquare(self.mofa_file)

    @staticmethod
    def get_factors(mofa_hdf5):
        factors_n = int(mofa_hdf5["training_stats"]["number_factors"][0])
        factors_labels = [f"F{i + 1}" for i in range(factors_n)]

        z = mofa_hdf5["expectations"]["Z"]
        factors = pd.concat(
            [
                pd.DataFrame(
                    df,
                    columns=[s.decode("utf-8") for s in mofa_hdf5["samples"][k]],
                    index=factors_labels,
                ).T
                for k, df in z.items()
            ]
        )
        return factors

    @staticmethod
    def get_weights(mofa_hdf5):
        factors_n = int(mofa_hdf5["training_stats"]["number_factors"][0])
        factors_labels = [f"F{i + 1}" for i in range(factors_n)]

        w = mofa_hdf5["expectations"]["W"]
        weights = {
            n: pd.DataFrame(
                w[n].value.T,
                index=[f.decode("utf-8") for f in mofa_hdf5["features"][n]],
                columns=factors_labels,
            )
            for n, df in w.items()
        }
        return weights

    @staticmethod
    def get_rsquare(mofa_hdf5):
        factors_n = int(mofa_hdf5["training_stats"]["number_factors"][0])
        factors_labels = [f"F{i + 1}" for i in range(factors_n)]

        r2 = mofa_hdf5["variance_explained"]["r2_per_factor"]
        rsquare = {
            k: pd.DataFrame(
                df,
                index=[s.decode("utf-8") for s in mofa_hdf5["views"]["views"]],
                columns=factors_labels,
            )
            for k, df in r2.items()
        }
        return rsquare

    @classmethod
    def read_mofa_hdf5(cls, hdf5_file):
        mofa_hdf5 = h5py.File(hdf5_file, "r")

        factors = cls.get_factors(mofa_hdf5)
        weights = cls.get_weights(mofa_hdf5)
        rsquare = cls.get_rsquare(mofa_hdf5)

        return factors, weights, rsquare

    @staticmethod
    def lm_residuals(y, x, fit_intercept=True, add_intercept=False):
        # Prepare input matrices
        ys = y.dropna()
        xs = x.loc[ys.index]
        xs = xs.loc[:, xs.std() > 0]

        if ys.shape[0] <= xs.shape[1]:
            return None

        # Linear regression models
        lm = LinearRegression(fit_intercept=fit_intercept).fit(xs, ys)

        # Calculate residuals
        residuals = ys - lm.predict(xs) - lm.intercept_

        # Add intercept
        if add_intercept:
            residuals += lm.intercept_

        return residuals

    def regress_out_covariates(self, fit_intercept=True, add_intercept=True):
        views_ = {}

        for k in self.views:
            if k in self.covariates:
                cov = self.covariates[k]

                # Confirm type is continuous
                k_dtype = self.views[k].dtypes[0]

                if k_dtype in [int]:
                    LOG.warning(
                        f"View {k} is of type {k_dtype}; covariates not regressed-out"
                    )
                    views_[k] = self.views[k].copy()
                    continue

                if self.verbose > 0:
                    LOG.info(
                        f"Regressing-out covariates (N={cov.shape}) from {k} view"
                    )

                # Regress-out
                views_[k] = pd.DataFrame(
                    {
                        i: self.lm_residuals(
                            self.views[k].loc[i],
                            cov,
                            fit_intercept=fit_intercept,
                            add_intercept=add_intercept,
                        )
                        for i in self.views[k].index
                    }
                ).T[self.samples]

            else:
                views_[k] = self.views[k].copy()

        return views_

    def pathway_enrichment(
        self, factor, views=None, genesets=None, nprocesses=4, permutation_num=0
    ):
        if genesets is None:
            genesets = [
                "c6.all.v7.1.symbols.gmt",
                "c5.all.v7.1.symbols.gmt",
                "h.all.v7.1.symbols.gmt",
                "c2.all.v7.1.symbols.gmt",
            ]

        if views is None:
            views = ["methylation", "transcriptomics", "proteomics"]

        df = pd.concat(
            [
                gseapy.ssgsea(
                    self.weights[v][factor],
                    processes=nprocesses,
                    permutation_num=permutation_num,
                    gene_sets=Enrichment.read_gmt(f"{DPATH}/pathways/{g}"),
                    no_plot=True,
                )
                .res2d.assign(geneset=g)
                .assign(view=v)
                .reset_index()
                for v in views
                for g in genesets
            ],
            ignore_index=True,
        )
        df = df.rename(columns={"sample1": "nes"}).sort_values("nes")
        return df

    def get_top_features(self, view, factor, n_features=30):
        df = self.weights[view][factor]
        df = df.loc[df.abs().sort_values(ascending=False).head(n_features).index]
        return df.sort_values()

    def get_factor_view_weights(self, factor):
        return pd.DataFrame({v: self.weights[v][factor] for v in self.weights})

    def save_hdf5(self, outfile):
        self.ep.save(outfile)


class MOFAPlot(CrispyPlot):
    @classmethod
    def factors_corr_clustermap(cls, mofa_obj, method="pearson", cmap="RdBu_r"):
        plot_df = mofa_obj.factors.corr(method=method)

        nrows, ncols = plot_df.shape
        fig = sns.clustermap(
            plot_df,
            cmap=cmap,
            center=0,
            annot=True,
            fmt=".2f",
            annot_kws={"size": 5},
            figsize=(min(0.3 * ncols, 10), min(0.3 * nrows, 10)),
        )
        fig.ax_heatmap.set_xlabel("Factors")
        fig.ax_heatmap.set_ylabel("Factors")

        fig.ax_heatmap.set_yticklabels(fig.ax_heatmap.get_yticklabels(), rotation=0)

    @classmethod
    def variance_explained_heatmap(cls, mofa_obj, row_order=None, rotate_ylabels=True):
        n_heatmaps = len(mofa_obj.rsquare)
        nrows, ncols = list(mofa_obj.rsquare.values())[0].shape
        row_order = (
            list(list(mofa_obj.rsquare.values())[0].index)
            if row_order is None
            else row_order
        )

        f, axs = plt.subplots(
            n_heatmaps,
            2,
            sharex="col",
            sharey="row",
            figsize=(0.3 * ncols, 0.5 * nrows),
            gridspec_kw=dict(width_ratios=[4, 1]),
        )

        for i, k in enumerate(mofa_obj.rsquare):
            axh, axb = axs[i][0] if n_heatmaps > 1 else axs[0], axs[i][1]

            df = mofa_obj.rsquare[k]

            # Heatmap
            g = sns.heatmap(
                df.loc[row_order],
                cmap="Greys",
                annot=True,
                cbar=False,
                fmt=".2f",
                linewidths=0.5,
                ax=axh,
                annot_kws={"fontsize": 5},
            )
            axh.set_xlabel("Factors" if i == (n_heatmaps - 1) else None)
            axh.set_ylabel(k)
            axh.set_title("Variance explained per factor" if i == 0 else None)
            g.set_yticklabels(
                g.get_yticklabels(), rotation=0 if rotate_ylabels else 90, horizontalalignment="right"
            )

            # Barplot
            i_sums = df.loc[row_order[::-1]].sum(1)

            norm = matplotlib.colors.Normalize(vmin=0, vmax=i_sums.max())
            colors = [plt.cm.Greys(norm(v)) for v in i_sums]

            axb.barh(np.arange(len(row_order)) + 0.5, i_sums, color=colors, linewidth=0)
            axb.set_xlabel("R2" if i == (n_heatmaps - 1) else None)
            axb.set_ylabel("")
            axb.set_title("Total variance per view" if i == 0 else None)
            axb.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0, axis="x")

        plt.subplots_adjust(hspace=0, wspace=0.05)

    @classmethod
    def factors_weights(cls, mofa_obj, view, cmap="RdYlGn"):
        nrows, ncols = mofa_obj.weights[view].shape
        fig = sns.clustermap(
            mofa_obj.weights[view],
            cmap=cmap,
            center=0,
            figsize=(0.25 * ncols, min(0.01 * nrows, 10)),
        )
        fig.ax_heatmap.set_xlabel("Factors")
        fig.ax_heatmap.set_ylabel(f"{view} features")

    @classmethod
    def factor_weights_scatter(
        cls, mofa_obj, view, factor, n_features=30, label_features=True, fontsize=4, r2=None
    ):
        df = mofa_obj.get_top_features(
            view, factor, n_features=n_features
        ).reset_index()

        fig, ax = plt.subplots(1, 1, figsize=(2.5, 2), dpi=600)

        ax.scatter(
            df.index,
            df[factor],
            marker="o",
            edgecolor="",
            s=5,
            alpha=0.8,
            c=cls.PAL_DTRACE[2],
        )

        if label_features:
            texts = [
                plt.text(i, w, g, color="k", fontsize=fontsize)
                for i, (g, w) in df.iterrows()
            ]
            adjust_text(
                texts, arrowprops=dict(arrowstyle="-", color="k", alpha=0.75, lw=0.3)
            )

        ax.set_ylabel(
            "Loading"
            if r2 is None
            else f"Loading ({mofa_obj.rsquare.loc[view, factor]:.2f}%)"
        )
        ax.set_xlabel("Rank position")
        ax.set_title(f"View: {view}; Factor: {factor}; Top {n_features} features")
        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)

        return ax

    @classmethod
    def view_heatmap(
        cls,
        mofa_obj,
        view,
        factor,
        df=None,
        cmap="Spectral",
        title="",
        center=None,
        col_colors=None,
        row_colors=None,
        mask=None,
        n_features=30,
        standard_scale=None,
    ):
        if df is None:
            df = mofa_obj.views[view]
            df = df.loc[
                mofa_obj.get_top_features(view, factor, n_features=n_features).index
            ]

        if mask is None:
            mask = df.isnull()

        if center is None:
            center = 0.5 if view == "methylation" else 0

        elif not center:
            center = None

        df = df.T.fillna(df.T.mean()).T

        nrows, ncols = mofa_obj.views[view].shape
        fig = sns.clustermap(
            df,
            mask=mask,
            cmap=cmap,
            center=center,
            standard_scale=standard_scale,
            col_colors=col_colors,
            row_colors=row_colors,
            xticklabels=False,
            figsize=(min(0.1 * ncols, 7), min(0.05 * nrows, 7)),
        )
        fig.ax_heatmap.set_xlabel("Samples")
        fig.ax_heatmap.set_ylabel(f"Features")
        fig.ax_col_dendrogram.set_title(title)

    @classmethod
    def covariates_heatmap(cls, covs, mofa_obj, agg):
        # Dendogram heatmap
        n_heatmaps = len(mofa_obj.rsquare)
        nrows, ncols = list(mofa_obj.rsquare.values())[0].shape

        row_order = list(list(mofa_obj.rsquare.values())[0].index)
        col_order = list(list(mofa_obj.rsquare.values())[0].columns)

        f, axs = plt.subplots(
            1,
            n_heatmaps + 2,
            sharex="none",
            sharey="none",

            gridspec_kw={"width_ratios": [0.4] * n_heatmaps + [9, 4]},
            figsize=(0.3 * n_heatmaps + 0.3 * covs.shape[0] + 0.3 * len(set(agg)), 0.3 * ncols),
        )

        vmax = np.max([mofa_obj.rsquare[k].max().max() for k in mofa_obj.rsquare])

        # Factors
        for i, k in enumerate(mofa_obj.rsquare):
            axh = axs[i]

            df = mofa_obj.rsquare[k]

            # Heatmap
            g = sns.heatmap(
                df.loc[row_order, col_order].T,
                cmap="Greys",
                annot=True,
                cbar=False,
                fmt=".2f",
                linewidths=0.5,
                ax=axh,
                vmin=0,
                vmax=vmax,
                annot_kws={"fontsize": 5},
            )
            axh.set_title(f"{k} cell lines")
            g.set_xticklabels(
                labels=row_order, rotation=90, horizontalalignment="center"
            )
            g.set_yticklabels(
                g.get_yticklabels() if i == 0 else [], rotation=0, horizontalalignment="right"
            )

            # # Barplot
            # i_sums = df.loc[row_order[::-1]].sum(1)
            #
            # norm = matplotlib.colors.Normalize(vmin=0, vmax=i_sums.max())
            # colors = [plt.cm.Greys(norm(v)) for v in i_sums]
            #
            # axb.barh(np.arange(len(row_order)) + 0.5, i_sums, color=colors, linewidth=0)
            # axb.set_xlabel("R2" if i == (n_heatmaps - 1) else None)
            # axb.set_ylabel("")
            # axb.set_title("Total variance per view" if i == 0 else None)
            # axb.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0, axis="x")

        # Covariates
        ax = axs[n_heatmaps]

        g = sns.heatmap(
            covs.T.loc[col_order],
            cmap="Spectral",
            center=0,
            annot=True,
            cbar=False,
            fmt=".2f",
            linewidths=0.5,
            ax=ax,
            annot_kws={"fontsize": 5},
        )
        g.set_yticklabels([])
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.set_title(f"Potential related factors")

        # Tissue contribution
        ax = axs[-1]

        g = sns.heatmap(
            mofa_obj.factors.groupby(agg).mean().T.loc[col_order],
            cmap=sns.diverging_palette(220, 20, sep=20, as_cmap=True),
            center=0,
            annot=True,
            cbar=False,
            fmt=".1f",
            linewidths=0.5,
            ax=ax,
            annot_kws={"fontsize": 5},
        )
        ax.yaxis.tick_right()
        g.set_yticklabels(
            g.get_yticklabels(), rotation=0, horizontalalignment="left"
        )
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.set_title(f"{agg.name} contribution")

        plt.subplots_adjust(wspace=0.01)

        # plt.suptitle("Variance explained per factor" if i == 0 else None)
        # axs[-1][1].axis('off')

        return axs



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
from crispy.CrispyPlot import CrispyPlot
from sklearn.decomposition import PCA, FactorAnalysis


def pc_labels(n):
    return [f"PC{i}" for i in np.arange(1, n + 1)]


def dim_reduction_pca(df, pca_ncomps=10):
    df_pca = PCA(n_components=pca_ncomps).fit(df.T)

    df_pcs = df_pca.transform(df.T)
    df_pcs = pd.DataFrame(df_pcs, index=df.T.index, columns=pc_labels(pca_ncomps))
    df_loadings = pd.DataFrame(df_pca.components_, index=pc_labels(pca_ncomps), columns=df.T.columns)

    df_vexp = pd.Series(df_pca.explained_variance_ratio_, index=df_pcs.columns)

    return df_pcs, df_vexp, df_loadings


def dim_reduction_fa(df, pca_ncomps=10):
    df_pca = FactorAnalysis(n_components=pca_ncomps).fit(df.T)

    df_pcs = df_pca.transform(df.T)
    df_pcs = pd.DataFrame(df_pcs, index=df.T.index, columns=pc_labels(pca_ncomps))
    df_loadings = pd.DataFrame(df_pca.components_, index=pc_labels(pca_ncomps), columns=df.T.columns)

    return df_pcs, df_loadings


def dim_reduction(
    df,
    pca_ncomps=50,
    tsne_ncomps=2,
    perplexity=30.0,
    early_exaggeration=12.0,
    learning_rate=200.0,
    n_iter=1000,
):
    # PCA
    df_pca = dim_reduction_pca(df, pca_ncomps)[0]

    # tSNE
    df_tsne = TSNE(
        n_components=tsne_ncomps,
        perplexity=perplexity,
        early_exaggeration=early_exaggeration,
        learning_rate=learning_rate,
        n_iter=n_iter,
    ).fit_transform(df_pca)
    df_tsne = pd.DataFrame(df_tsne, index=df_pca.index, columns=pc_labels(tsne_ncomps))

    return df_tsne, df_pca


def plot_dim_reduction(data, palette=None, ctype="tSNE", hue_filed="tissue"):
    if hue_filed not in data.columns:
        data = data.assign(tissue="All")

    if palette is None:
        palette = dict(All=CrispyPlot.PAL_DBGD[0])

    fig, ax = plt.subplots(1, 1, figsize=(4.0, 4.0), dpi=600)

    for t, df in data.groupby(hue_filed):
        ax.scatter(
            df["PC1"],
            df["PC2"],
            c=palette[t],
            marker="o",
            edgecolor="",
            s=5,
            label=t,
            alpha=0.8,
        )
    ax.set_xlabel("Dimension 1" if ctype == "tSNE" else "PC 1")
    ax.set_ylabel("Dimension 2" if ctype == "tSNE" else "PC 2")
    ax.axis("off" if ctype == "tSNE" else "on")

    if ctype == "pca":
        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)

    ax.legend(
        loc="center left",
        bbox_to_anchor=(1, 0.5),
        prop={"size": 4},
        frameon=False,
        title=hue_filed,
    ).get_title().set_fontsize("5")

    return ax



#!/usr/bin/env python
# Copyright (C) 2018 Emanuel Goncalves

import crispy as cy
import matplotlib.pyplot as plt

# Import data
rawcounts, copynumber = cy.Utils.get_example_data()

# Import CRISPR-Cas9 library
lib = cy.Utils.get_crispr_lib()

# Instantiate Crispy
crispy = cy.Crispy(
    raw_counts=rawcounts, copy_number=copynumber, library=lib
)

# Fold-changes and correction integrated funciton.
# Output is a modified/expanded BED output with sgRNA and segments information
bed_df = crispy.correct(x_features='ratio', y_feature='fold_change')
print(bed_df.head())

# Gaussian Process Regression is stored
crispy.gpr.plot(x_feature='ratio', y_feature='fold_change')
plt.show()



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import logging
import numpy as np
import pandas as pd
import pkg_resources
import itertools as it
from crispy.Utils import Utils
from scipy.stats import shapiro, iqr
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import quantile_transform
from statsmodels.stats.multitest import multipletests


LOG = logging.getLogger("Crispy")
DPATH = pkg_resources.resource_filename("crispy", "data/")


class Sample:
    """
    Import module that handles the sample list (i.e. list of cell lines) and their descriptive information.

    """

    def __init__(
        self,
        index="model_id",
        samplesheet_file="model_list_20200204.csv",
        growth_file="growth_rates_rapid_screen_1536_v1.6.3_02Jun20.csv",
        medium_file="SIDMvsMedia.xlsx",
        institute_file="crispr/CRISPR_Institute_Origin_20191108.csv.gz",
    ):
        self.index = index

        # Import samplesheet
        self.samplesheet = (
            pd.read_csv(f"{DPATH}/{samplesheet_file}")
            .dropna(subset=[self.index])
            .set_index(self.index)
        )

        # Growth rates
        self.growth = pd.read_csv(f"{DPATH}/{growth_file}")
        self.samplesheet["growth"] = (
            self.growth.groupby(self.index)["GROWTH_RATE"]
            .mean()
            .reindex(self.samplesheet.index)
            .values
        )

        # CRISPR institute
        self.samplesheet["institute"] = (
            pd.read_csv(f"{DPATH}/{institute_file}", index_col=0, header=None)
            .iloc[:, 0]
            .reindex(self.samplesheet.index)
            .values
        )

        # Breakdown tissue type
        self.samplesheet["model_type"] = [
            c if t in ["Lung", "Haematopoietic and Lymphoid"] else t
            for t, c in self.samplesheet[["tissue", "cancer_type"]].values
        ]

        # Screen medium
        self.media = pd.read_excel(f"{DPATH}/{medium_file}")
        self.media = self.media.groupby("SIDM")["Screen Media"].first()
        self.samplesheet["media"] = self.media.reindex(self.samplesheet.index)

    def get_covariates(self, culture_conditions=True, cancer_type=True):
        covariates = []

        # Cell lines culture conditions
        if culture_conditions:
            culture = pd.get_dummies(self.samplesheet["growth_properties"]).drop(
                columns=["Unknown"]
            )
            covariates.append(culture)

        # Cell lines culture conditions
        if cancer_type:
            ctype = pd.get_dummies(self.samplesheet["cancer_type"])
            covariates.append(ctype)

        # Merge covariates
        covariates = pd.concat(covariates, axis=1, sort=False)

        return covariates


class WES:
    def __init__(self, wes_file="WES_variants.csv.gz"):
        self.wes = pd.read_csv(f"{DPATH}/wes/{wes_file}")

    def get_data(self, as_matrix=True, mutation_class=None, recurrence=False):
        df = self.wes.copy()

        # Filter by mutation types
        if mutation_class is not None:
            df = df[df["Classification"].isin(mutation_class)]

        if recurrence:
            df = df[df["Recurrence Filter"] == "Yes"]

        if as_matrix:
            df["value"] = 1

            df = pd.pivot_table(
                df,
                index="Gene",
                columns="model_id",
                values="value",
                aggfunc="first",
                fill_value=0,
            )

        return df

    def filter(
        self,
        subset=None,
        min_events=5,
        as_matrix=True,
        mutation_class=None,
        recurrence=False,
    ):
        df = self.get_data(
            as_matrix=as_matrix, mutation_class=mutation_class, recurrence=recurrence
        )

        # Subset samples
        if subset is not None:
            if as_matrix:
                df = df.loc[:, df.columns.isin(subset)]

            else:
                df = df[df["model_id"].isin(subset)]

            assert df.shape[1] != 0, "No columns after filter by subset"

        # Minimum number of events
        if min_events is not None:
            df = df[df.sum(1) >= min_events]

        return df


class GeneExpression:
    """
    Import module of gene-expression data-set.

    """

    def __init__(
        self,
        voom_file="gexp/rnaseq_voom.csv.gz",
        read_count="gexp/rnaseq_20191101/rnaseq_read_count_20191101.csv",
    ):
        self.voom = pd.read_csv(f"{DPATH}/{voom_file}", index_col=0)
        self.readcount = pd.read_csv(f"{DPATH}/{read_count}", index_col=1).drop(
            columns=["model_id"]
        )
        self.discrete = pd.read_csv(
            f"{DPATH}/GDSC_discretised_table.csv.gz", index_col=0
        )

    def get_data(self, dtype="voom"):
        if dtype == "voom":
            return self.voom.copy()

        elif dtype == "readcount":
            return self.readcount.copy()

        else:
            assert False, f"Dtype {dtype} not supported"

    def filter(
        self,
        dtype="voom",
        subset=None,
        iqr_range=None,
        normality=False,
        lift_gene_ids=True,
    ):
        df = self.get_data(dtype=dtype)

        # Subset matrices
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        # Filter by IQR
        if iqr_range is not None:
            iqr_ranges = (
                df.apply(lambda v: iqr(v, rng=iqr_range), axis=1)
                .rename("iqr")
                .to_frame()
            )

            gm_iqr = GaussianMixture(n_components=2).fit(iqr_ranges[["iqr"]])
            iqr_ranges["gm"] = gm_iqr.predict(iqr_ranges[["iqr"]])

            df = df.reindex(iqr_ranges["gm"] != gm_iqr.means_.argmin())

            LOG.info(f"IQR {iqr_range}")
            LOG.info(iqr_ranges.groupby("gm").agg({"min", "mean", "median", "max"}))

        # Filter by normality
        if normality:
            normality = df.apply(lambda v: shapiro(v)[1], axis=1)
            normality = multipletests(normality, method="bonferroni")
            df = df.reindex(normality[0])

        if lift_gene_ids:
            gmap = (
                pd.read_csv(f"{DPATH}/gexp/hgnc-symbol-check.csv")
                .groupby("Input")["Approved symbol"]
                .first()
            )
            df.index = gmap.loc[df.index]

        return df

    def is_not_expressed(self, rpkm_threshold=1, subset=None):
        rpkm = self.filter(dtype="rpkm", subset=subset)
        rpkm = (rpkm < rpkm_threshold).astype(int)
        return rpkm


class DrugResponse:
    """
    Importer module for drug-response measurements acquired at Sanger Institute GDSC (https://cancerrxgene.org).

    """

    SAMPLE_COLUMNS = ["model_id"]
    DRUG_COLUMNS = ["drug_id", "drug_name", "dataset"]

    def __init__(
        self,
        drugresponse_file="drugresponse/DrugResponse_PANCANCER_GDSC1_GDSC2_20200602.csv.gz",
    ):
        # Import and Merge drug response matrix (IC50)
        self.drugresponse = pd.read_csv(f"{DPATH}/{drugresponse_file}")
        self.drugresponse = self.drugresponse[
            ~self.drugresponse["cell_line_name"].isin(["LS-1034"])
        ]

        # Drug max concentration
        self.maxconcentration = self.drugresponse.groupby(self.DRUG_COLUMNS)[
            "max_screening_conc"
        ].first()

    @staticmethod
    def assemble():
        gdsc1 = pd.read_csv(
            f"{DPATH}/drugresponse/fitted_data_screen_96_384_v1.6.0_02Jun20.csv"
        )
        gdsc1 = gdsc1.assign(dataset="GDSC1").query("(RMSE < 0.3)")
        gdsc1 = gdsc1.query("use_in_publications == 'Y'")

        gdsc2 = pd.read_csv(
            f"{DPATH}/drugresponse/fitted_data_rapid_screen_1536_v1.6.3_02Jun20.csv"
        )
        gdsc2 = gdsc2.assign(dataset="GDSC2").query("(RMSE < 0.3)")
        gdsc2 = gdsc2.query("use_in_publications == 'Y'")

        columns = set(gdsc1).intersection(gdsc2)

        drespo = pd.concat([gdsc1[columns], gdsc2[columns]], axis=0, ignore_index=True)

        drespo.to_csv(
            f"{DPATH}/drugresponse/DrugResponse_PANCANCER_GDSC1_GDSC2_20200602.csv.gz",
            compression="gzip",
            index=False,
        )

    def get_data(self, dtype="ln_IC50"):
        data = pd.pivot_table(
            self.drugresponse,
            index=self.DRUG_COLUMNS,
            columns=self.SAMPLE_COLUMNS,
            values=dtype,
            fill_value=np.nan,
        )

        return data

    def filter(
        self,
        dtype="ln_IC50",
        subset=None,
        min_events=3,
        min_meas=25,
        max_c=0.5,
        filter_min_observations=False,
        filter_max_concentration=False,
        filter_combinations=False,
    ):
        # Drug max screened concentration
        df = self.get_data(dtype="ln_IC50")
        d_maxc = np.log(self.maxconcentration * max_c)

        # - Filters
        # Subset samples
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        # Filter by mininum number of observations
        if filter_min_observations:
            df = df[df.count(1) > min_meas]

        # Filter by max screened concentration
        if filter_max_concentration:
            df = df[[sum(df.loc[i] < d_maxc.loc[i]) >= min_events for i in df.index]]

        # Filter combinations
        if filter_combinations:
            df = df[[" + " not in i[1] for i in df.index]]

        return self.get_data(dtype=dtype).loc[df.index, df.columns]


class BioGRID:
    def __init__(
        self,
        biogrid_file="BIOGRID-ALL-3.5.180.tab2.zip",
        organism=9606,
        etype="physical",
        stypes_exclude=None,
        homodymers_exclude=True,
        ddir=None,
    ):
        self.ddir = DPATH if ddir is None else ddir

        self.etype = etype
        self.organism = organism
        self.homodymers_exclude = homodymers_exclude
        self.stypes_exclude = (
            {"Affinity Capture-RNA", "Protein-RNA"}
            if stypes_exclude is None
            else stypes_exclude
        )

        # Import
        self.biogrid = pd.read_csv(f"{self.ddir}/{biogrid_file}", sep="\t")

        # Filter by organism
        self.biogrid = self.biogrid[
            self.biogrid["Organism Interactor A"] == self.organism
        ]
        self.biogrid = self.biogrid[
            self.biogrid["Organism Interactor B"] == self.organism
        ]

        # Filter by type of interaction
        if self.etype is not None:
            self.biogrid = self.biogrid[
                self.biogrid["Experimental System Type"] == self.etype
            ]

        # Exlude experimental systems
        self.biogrid = self.biogrid[
            ~self.biogrid["Experimental System"].isin(self.stypes_exclude)
        ]

        # Exclude homodymers
        if self.homodymers_exclude:
            self.biogrid = self.biogrid[
                self.biogrid["Official Symbol Interactor A"]
                != self.biogrid["Official Symbol Interactor B"]
            ]

        # Build set of interactions (both directions, i.e. p1-p2, p2-p1)
        self.biogrid = {
            (p1, p2)
            for p1, p2 in self.biogrid[
                ["Official Symbol Interactor A", "Official Symbol Interactor B"]
            ].values
        }
        self.biogrid = {
            (p1, p2) for p in self.biogrid for p1, p2 in [(p[0], p[1]), (p[1], p[0])]
        }


class PPI:
    """
    Module used to import protein-protein interaction network

    """

    def __init__(
        self,
        string_file="9606.protein.links.full.v11.0.txt.gz",
        string_alias_file="9606.protein.aliases.v11.0.txt.gz",
        ddir=None,
    ):
        self.ddir = DPATH if ddir is None else ddir

        self.string_file = string_file
        self.string_alias_file = string_alias_file

    @classmethod
    def ppi_annotation(
        cls, df, ppi, target_thres=5, y_var="y_id", x_var="x_id", ppi_var="x_ppi"
    ):
        genes_source = set(
            {g for v in df[y_var].dropna() for g in v.split(";")}
        ).intersection(set(ppi.vs["name"]))
        genes_target = set(
            {g for v in df[x_var].dropna() for g in v.split(";")}
        ).intersection(set(ppi.vs["name"]))

        # Calculate distance between drugs and genes in PPI
        dist_g_g = {
            g: pd.Series(
                ppi.shortest_paths(source=g, target=genes_target)[0], index=genes_target
            ).to_dict()
            for g in genes_source
        }

        def gene_gene_annot(g_source, g_target):
            if str(g_source) == "nan" or str(g_target) == "nan":
                res = np.nan

            elif len(set(g_source.split(";")).intersection(g_target.split(";"))) > 0:
                res = "T"

            elif g_source not in genes_source:
                res = "-"

            elif g_target not in genes_target:
                res = "-"

            else:
                g_st_min = np.min(
                    [
                        dist_g_g[gs][gt]
                        for gs in g_source.split(";")
                        for gt in g_target.split(";")
                    ]
                )
                res = cls.ppi_dist_to_string(g_st_min, target_thres)

            return res

        # Annotate drug regressions
        df = df.assign(
            x_ppi=[
                gene_gene_annot(g_source, g_target)
                for g_source, g_target in df[[y_var, x_var]].values
            ]
        )
        df = df.rename(columns=dict(x_ppi=ppi_var))

        return df

    @staticmethod
    def ppi_dist_to_string(d, target_thres):
        if d == 0:
            res = "T"

        elif d == np.inf:
            res = "-"

        elif d < target_thres:
            res = f"{int(d)}"

        else:
            res = f"{int(target_thres)}+"

        return res

    def build_string_ppi(self, score_thres=900, export_pickle=None):
        import igraph

        # ENSP map to gene symbol
        gmap = pd.read_csv(f"{self.ddir}/{self.string_alias_file}", sep="\t")
        gmap = gmap[["BioMart_HUGO" in i.split(" ") for i in gmap["source"]]]
        gmap = (
            gmap.groupby("string_protein_id")["alias"].agg(lambda x: set(x)).to_dict()
        )
        gmap = {k: list(gmap[k])[0] for k in gmap if len(gmap[k]) == 1}
        logging.getLogger("DTrace").info(f"ENSP gene map: {len(gmap)}")

        # Load String network
        net = pd.read_csv(f"{self.ddir}/{self.string_file}", sep=" ")

        # Filter by moderate confidence
        net = net[net["combined_score"] > score_thres]

        # Filter and map to gene symbol
        net = net[
            [
                p1 in gmap and p2 in gmap
                for p1, p2 in net[["protein1", "protein2"]].values
            ]
        ]
        net["protein1"] = [gmap[p1] for p1 in net["protein1"]]
        net["protein2"] = [gmap[p2] for p2 in net["protein2"]]
        LOG.info(f"String: {len(net)}")

        #  String network
        net_i = igraph.Graph(directed=False)

        # Initialise network lists
        edges = [(px, py) for px, py in net[["protein1", "protein2"]].values]
        vertices = list(set(net["protein1"]).union(net["protein2"]))

        # Add nodes
        net_i.add_vertices(vertices)

        # Add edges
        net_i.add_edges(edges)

        # Add edge attribute score
        net_i.es["score"] = list(net["combined_score"])

        # Simplify
        net_i = net_i.simplify(combine_edges="max")
        LOG.info(net_i.summary())

        # Export
        if export_pickle is not None:
            net_i.write_pickle(export_pickle)

        return net_i

    @staticmethod
    def ppi_corr(ppi, m_corr, m_corr_thres=None):
        """
        Annotate PPI network based on Pearson correlation between the vertices of each edge using
        m_corr data-frame and m_corr_thres (Pearson > m_corr_thress).

        :param ppi:
        :param m_corr:
        :param m_corr_thres:
        :return:
        """
        # Subset PPI network
        ppi = ppi.subgraph([i.index for i in ppi.vs if i["name"] in m_corr.index])

        # Edge correlation
        crispr_pcc = np.corrcoef(m_corr.loc[ppi.vs["name"]].values)
        ppi.es["corr"] = [crispr_pcc[i.source, i.target] for i in ppi.es]

        # Sub-set by correlation between vertices of each edge
        if m_corr_thres is not None:
            ppi = ppi.subgraph_edges(
                [i.index for i in ppi.es if abs(i["corr"]) > m_corr_thres]
            )

        LOG.info(ppi.summary())

        return ppi

    @classmethod
    def get_edges(cls, ppi, nodes, corr_thres, norder):
        # Subset network
        ppi_sub = ppi.copy().subgraph_edges(
            [e for e in ppi.es if abs(e["corr"]) >= corr_thres]
        )

        # Nodes that are contained in the network
        nodes = {v for v in nodes if v in ppi_sub.vs["name"]}
        assert len(nodes) > 0, "None of the nodes is contained in the PPI"

        # Nodes neighborhood
        neighbor_nodes = {
            v for n in nodes for v in ppi_sub.neighborhood(n, order=norder)
        }

        # Build subgraph
        subgraph = ppi_sub.subgraph(neighbor_nodes)

        # Build data-frame
        nodes_df = pd.DataFrame(
            [
                {
                    "source": subgraph.vs[e.source]["name"],
                    "target": subgraph.vs[e.target]["name"],
                    "r": e["corr"],
                }
                for e in subgraph.es
            ]
        ).sort_values("r")

        return nodes_df


class CORUM:
    def __init__(
        self,
        corum_file="coreComplexes.txt",
        organism="Human",
        homodymers_exclude=True,
        protein_subset=None,
        ddir=None,
    ):
        self.ddir = DPATH if ddir is None else ddir

        self.organism = organism
        self.homodymers_exclude = homodymers_exclude
        self.protein_subset = protein_subset

        # Load CORUM DB
        self.db = pd.read_csv(f"{self.ddir}/{corum_file}", sep="\t")
        self.db = self.db.query(f"Organism == '{organism}'")
        self.db_name = self.db.groupby("ComplexID")["ComplexName"].first()

        # Melt into list of protein pairs (both directions, i.e. p1-p2, p2-p1)
        self.db_melt = self.melt_ppi()

        # Map to gene symbols
        self.gmap = self.map_gene_name()
        self.db_melt_symbol = {
            (self.gmap.loc[p1, "GeneSymbol"], self.gmap.loc[p2, "GeneSymbol"]): i
            for (p1, p2), i in self.db_melt.items()
            if p1 in self.gmap.index and p2 in self.gmap.index
        }

        # Exclude homodymers
        if self.homodymers_exclude:
            self.db_melt_symbol = {
                (p1, p2): i for (p1, p2), i in self.db_melt_symbol.items() if p1 != p2
            }

        # Subset interactions
        if self.protein_subset is not None:
            self.db_melt_symbol = {
                (p1, p2): i
                for (p1, p2), i in self.db_melt_symbol.items()
                if p1 in self.protein_subset and p2 in self.protein_subset
            }

    def melt_ppi(self, idx_id="ComplexID", idx_sub="subunits(UniProt IDs)"):
        db_melt = self.db[[idx_id, idx_sub]].copy()
        db_melt[idx_sub] = db_melt[idx_sub].apply(
            lambda v: list(it.permutations(v.split(";"), 2))
        )
        db_melt = {p: i for i, c in db_melt[[idx_id, idx_sub]].values for p in c}
        return db_melt

    def map_gene_name(self, index_col="Entry"):
        idmap = pd.read_csv(f"{self.ddir}/uniprot_human_idmap.tab.gz", sep="\t")

        if index_col is not None:
            idmap = idmap.dropna(subset=[index_col]).set_index(index_col)

        idmap["GeneSymbol"] = idmap["Gene names  (primary )"].apply(
            lambda v: v.split("; ")[0] if str(v).lower() != "nan" else v
        )

        return idmap


class HuRI:
    def __init__(self, ppi_file="HuRI.tsv", idmap_file="HuRI_biomart_idmap.tsv", ddir=None):
        self.ddir = DPATH if ddir is None else ddir

        self.huri = pd.read_csv(f"{self.ddir}/{ppi_file}", sep="\t", header=None)

        # Convert to a set of pairs {(p1, p2), ...}
        self.huri = {(p1, p2) for p1, p2 in self.huri.values}

        # Map ids
        idmap = pd.read_csv(f"{self.ddir}/{idmap_file}", sep="\t", index_col=0)[
            "Gene name"
        ].to_dict()
        self.huri = {
            (idmap[p1], idmap[p2])
            for p1, p2 in self.huri
            if p1 in idmap and p2 in idmap
        }

        # Remove self interactions
        self.huri = {(p1, p2) for p1, p2 in self.huri if p1 != p2}

        # Build set of interactions (both directions, i.e. p1-p2, p2-p1)
        self.huri = {
            (p1, p2) for p in self.huri for p1, p2 in [(p[0], p[1]), (p[1], p[0])]
        }


class Metabolomics:
    def __init__(self, metab_file="metabolomics/CCLE_metabolomics_20190502.csv"):
        m_ss = Sample().samplesheet
        m_ss = m_ss.reset_index().dropna(subset=["BROAD_ID"]).set_index("BROAD_ID")

        # Import
        self.metab = pd.read_csv(f"{DPATH}/{metab_file}")
        self.metab["model_id"] = self.metab["DepMap_ID"].replace(m_ss["model_id"])
        self.metab = self.metab.groupby("model_id").mean().T

    def get_data(self):
        return self.metab.copy()

    def filter(
        self,
        dtype="protein",
        subset=None,
        normality=False,
        iqr_range=None,
        perc_measures=None,
        quantile_normalise=False,
    ):
        df = self.get_data()

        # Subset matrices
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        return df


class Proteomics:
    def __init__(
        self,
        protein_matrix="proteomics/E0022_P06_Protein_Matrix_ProNorM.tsv.gz",
        protein_raw_matrix="proteomics/E0022_P06_Protein_Matrix_Raw_Mean_Intensities.tsv.gz",
        protein_mean_raw="proteomics/E0022_P06_Protein_Mean_Raw_Intensities.tsv",
        protein_rep_corr="proteomics/E0022_P06_final_reps_correlation.csv",
        manifest="proteomics/E0022_P06_final_sample_map.txt",
        samplesheet="proteomics/E0022_P06_samplehseet.csv",
        broad_tmt="proteomics/broad_tmt.csv.gz",
        coread_tmt="proteomics/proteomics_coread_processed_parsed.csv",
        hgsc_prot="proteomics/hgsc_cell_lines_proteomics.csv",
        brca_prot="proteomics/brca_cell_lines_proteomics_preprocessed.csv",
    ):
        self.ss = pd.read_csv(f"{DPATH}/{samplesheet}", index_col=0)

        deprecated_ids = self.map_deprecated()

        # Import manifest
        self.manifest = pd.read_csv(f"{DPATH}/{manifest}", index_col=0, sep="\t")

        # Remove excluded samples
        self.exclude_man = self.manifest[~self.manifest["SIDM"].isin(self.ss.index)]
        self.manifest = self.manifest[~self.manifest.index.isin(self.exclude_man.index)]

        # Replicate correlation
        self.reps = pd.read_csv(f"{DPATH}/{protein_rep_corr}", index_col=0).iloc[:, 0]

        # Import mean protein abundance
        self.protein_raw = pd.read_csv(
            f"{DPATH}/{protein_raw_matrix}", sep="\t", index_col=0
        )
        self.peptide_raw_mean = pd.read_csv(
            f"{DPATH}/{protein_mean_raw}", sep="\t", index_col=0
        ).iloc[:, 0]

        # Import imputed protein levels
        self.protein = pd.read_csv(f"{DPATH}/{protein_matrix}", sep="\t", index_col=0).T

        self.protein["Protein"] = (
            self.protein.reset_index()["index"]
            .replace(deprecated_ids["Entry name"])
            .values
        )
        self.protein = self.protein.set_index("Protein")
        self.protein = self.protein.rename(
            columns=self.manifest.groupby("Cell_line")["SIDM"].first()
        )

        exclude_controls = [
            "Control_HEK293T_lys",
            "Control_HEK293T_std_H002",
            "Control_HEK293T_std_H003",
        ]
        self.protein = self.protein.drop(columns=exclude_controls)

        # Import Broad TMT data-set
        self.broad = pd.read_csv(f"{DPATH}/{broad_tmt}", compression="gzip")
        self.broad = (
            self.broad.dropna(subset=["Gene_Symbol"])
            .groupby("Gene_Symbol")
            .agg(np.nanmean)
        )

        # Import CRC COREAD TMT
        self.coread = pd.read_csv(f"{DPATH}/{coread_tmt}", index_col=0)
        self.coread = self.coread.loc[
            :, self.coread.columns.isin(self.ss["model_name"])
        ]

        coread_ss = self.ss[self.ss["model_name"].isin(self.coread.columns)]
        coread_ss = coread_ss.reset_index().set_index("model_name")
        self.coread = self.coread.rename(columns=coread_ss["model_id"])

        # Import HGSC proteomics
        self.hgsc = (
            pd.read_csv(f"{DPATH}/{hgsc_prot}")
            .dropna(subset=["Gene names"])
            .drop(columns=["Majority protein IDs"])
        )
        self.hgsc = self.hgsc.groupby("Gene names").mean()
        self.hgsc = self.hgsc.loc[:, self.hgsc.columns.isin(self.ss["model_name"])]

        hgsc_ss = self.ss[self.ss["model_name"].isin(self.hgsc.columns)]
        hgsc_ss = hgsc_ss.reset_index().set_index("model_name")
        self.hgsc = self.hgsc.rename(columns=hgsc_ss["model_id"])

        # Import BRCA proteomics
        self.brca = pd.read_csv(f"{DPATH}/{brca_prot}", index_col=0)
        self.brca = self.brca.loc[:, self.brca.columns.isin(self.ss["model_name"])]

        brca_ss = self.ss[self.ss["model_name"].isin(self.brca.columns)]
        brca_ss = brca_ss.reset_index().set_index("model_name")
        self.brca = self.brca.rename(columns=brca_ss["model_id"])

    def get_data(self, dtype="protein", map_ids=True, quantile_normalise=False):
        if dtype.lower() == "protein":
            data = self.protein.copy()

        else:
            assert False, f"{dtype} not supported"

        if quantile_normalise:
            data = pd.DataFrame(
                quantile_transform(data, ignore_implicit_zeros=True),
                index=data.index,
                columns=data.columns,
            )

        if map_ids:
            pmap = self.map_gene_name().reindex(data.index)["GeneSymbol"].dropna()

            data = data[data.index.isin(pmap.index)]
            data = data.groupby(pmap.reindex(data.index)).mean()

        return data

    def filter(
        self,
        dtype="protein",
        subset=None,
        normality=False,
        iqr_range=None,
        perc_measures=None,
        quantile_normalise=False,
    ):
        df = self.get_data(dtype=dtype, quantile_normalise=quantile_normalise)

        # Subset matrices
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        # Filter by IQR
        if iqr_range is not None:
            iqr_ranges = (
                df.apply(lambda v: iqr(v, rng=iqr_range), axis=1)
                .rename("iqr")
                .to_frame()
            )

            gm_iqr = GaussianMixture(n_components=2).fit(iqr_ranges[["iqr"]])
            iqr_ranges["gm"] = gm_iqr.predict(iqr_ranges[["iqr"]])

            df = df.reindex(iqr_ranges["gm"] != gm_iqr.means_.argmin())

            LOG.info(f"IQR {iqr_range}")
            LOG.info(iqr_ranges.groupby("gm").agg({"min", "mean", "median", "max"}))

        # Filter by normality
        if normality:
            normality = df.apply(lambda v: shapiro(v)[1], axis=1)
            normality = multipletests(normality, method="bonferroni")
            df = df.reindex(normality[0])

        # Filter by number of obvservations
        if perc_measures is not None:
            df = df[df.count(1) > (perc_measures * df.shape[1])]

        return df

    @staticmethod
    def map_deprecated():
        return pd.read_csv(
            f"{DPATH}/uniprot_human_idmap_deprecated.tab", sep="\t", index_col=0
        )

    @staticmethod
    def map_gene_name(index_col="Entry name"):
        idmap = pd.read_csv(f"{DPATH}/uniprot_human_idmap.tab.gz", sep="\t")

        if index_col is not None:
            idmap = idmap.dropna(subset=[index_col]).set_index(index_col)

        idmap["GeneSymbol"] = idmap["Gene names  (primary )"].apply(
            lambda v: v.split("; ")[0] if str(v).lower() != "nan" else v
        )

        return idmap

    def calculate_mean_protein_intensities(
        self, peptide_matrix_raw="proteomics/E0022_P06_Peptide_Matrix_Raw.tsv.gz"
    ):
        peptide_raw = pd.read_csv(
            f"{DPATH}/{peptide_matrix_raw}", sep="\t", index_col=0
        ).T

        peptide_raw_mean = (
            peptide_raw.pipe(np.log2).groupby(self.manifest["SIDM"], axis=1).mean()
        )
        peptide_raw_mean = peptide_raw_mean.groupby(
            [p.split("=")[0] for p in peptide_raw_mean.index]
        ).mean()
        peptide_raw_mean = peptide_raw_mean.mean(1).sort_values()

        pmap = (
            self.map_gene_name().reindex(peptide_raw_mean.index)["GeneSymbol"].dropna()
        )

        peptide_raw_mean = peptide_raw_mean[peptide_raw_mean.index.isin(pmap.index)]
        peptide_raw_mean = peptide_raw_mean.groupby(
            pmap.reindex(peptide_raw_mean.index)
        ).mean()

        return peptide_raw_mean

    def replicates_correlation(
        self, reps_file="proteomics/E0022_P06_Protein_Matrix_Replicate_ProNorM.tsv.gz"
    ):
        reps = pd.read_csv(f"{DPATH}/{reps_file}", sep="\t", index_col=0).T

        reps_corr = {}

        for n, df in reps.groupby(self.manifest["SIDM"], axis=1):
            df_corr = df.corr()
            df_corr = pd.DataFrame(df_corr.pipe(np.triu, k=1)).replace(0, np.nan)
            reps_corr[n] = df_corr.unstack().dropna().mean()

        reps_corr = pd.Series(reps_corr, name="RepsCorrelation").sort_values(
            ascending=False
        )

        return reps_corr


class CRISPR:
    """
    Importer module for CRISPR-Cas9 screens acquired at Sanger and Broad Institutes.

    """

    def __init__(
        self,
        fc_file="crispr/CRISPR_corrected_qnorm_20191108.csv.gz",
        institute_file="crispr/CRISPR_Institute_Origin_20191108.csv.gz",
        merged_file="crispr/CRISPRcleanR_FC.txt.gz",
    ):
        self.crispr = pd.read_csv(f"{DPATH}/{fc_file}", index_col=0)
        self.institute = pd.read_csv(
            f"{DPATH}/{institute_file}", index_col=0, header=None
        ).iloc[:, 0]

        sid = (
            Sample()
            .samplesheet.reset_index()
            .dropna(subset=["BROAD_ID"])
            .groupby("BROAD_ID")["model_id"]
            .first()
        )

        self.merged = pd.read_csv(f"{DPATH}/{merged_file}", index_col=0, sep="\t")
        self.merged_institute = pd.Series(
            {c: "Broad" if c.startswith("ACH-") else "Sanger" for c in self.merged}
        )

        self.merged = self.merged.rename(columns=sid)
        self.merged_institute = self.merged_institute.rename(index=sid)

    def get_data(self, scale=True, dtype="merged"):
        if dtype == "merged":
            df = self.merged.copy()
        else:
            df = self.crispr.copy()

        if scale:
            df = self.scale(df)

        return df

    def filter(
        self,
        dtype="merged",
        subset=None,
        scale=True,
        std_filter=False,
        abs_thres=None,
        drop_core_essential=False,
        min_events=5,
        drop_core_essential_broad=False,
        binarise_thres=None,
    ):
        df = self.get_data(scale=True, dtype=dtype)

        # - Filters
        # Subset matrices
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        # Filter by scaled scores
        if abs_thres is not None:
            df = df[(df.abs() > abs_thres).sum(1) >= min_events]

        # Filter out core essential genes
        if drop_core_essential:
            df = df[~df.index.isin(Utils.get_adam_core_essential())]

        if drop_core_essential_broad:
            df = df[~df.index.isin(Utils.get_broad_core_essential())]

        # - Subset matrices
        x = self.get_data(scale=scale, dtype=dtype).reindex(
            index=df.index, columns=df.columns
        )

        if binarise_thres is not None:
            x = (x < binarise_thres).astype(int)

        if std_filter:
            x = x.reindex(x.std(1) > 0)

        return x

    @staticmethod
    def scale(df, essential=None, non_essential=None, metric=np.median):
        if essential is None:
            essential = Utils.get_essential_genes(return_series=False)

        if non_essential is None:
            non_essential = Utils.get_non_essential_genes(return_series=False)

        assert (
            len(essential.intersection(df.index)) != 0
        ), "DataFrame has no index overlapping with essential list"

        assert (
            len(non_essential.intersection(df.index)) != 0
        ), "DataFrame has no index overlapping with non essential list"

        essential_metric = metric(df.reindex(essential).dropna(), axis=0)
        non_essential_metric = metric(df.reindex(non_essential).dropna(), axis=0)

        df = df.subtract(non_essential_metric).divide(
            non_essential_metric - essential_metric
        )

        return df


class Mobem:
    """
    Import module for Genomic binary feature table (containing mutations and copy-number calls)
    Iorio et al., Cell, 2016.

    """

    def __init__(
        self, mobem_file="mobem/PANCAN_mobem.csv.gz", drop_factors=True, add_msi=True
    ):
        self.sample = Sample()

        idmap = (
            self.sample.samplesheet.reset_index()
            .dropna(subset=["COSMIC_ID", "model_id"])
            .set_index("COSMIC_ID")["model_id"]
        )

        mobem = pd.read_csv(f"{DPATH}/{mobem_file}", index_col=0)
        mobem = mobem[mobem.index.astype(str).isin(idmap.index)]
        mobem = mobem.set_index(idmap[mobem.index.astype(str)].values)

        if drop_factors is not None:
            mobem = mobem.drop(columns={"TISSUE_FACTOR", "MSI_FACTOR", "MEDIA_FACTOR"})

        if add_msi:
            self.msi = self.sample.samplesheet.reindex(mobem.index)["msi_status"]
            mobem["msi_status"] = (self.msi == "MSI-H").astype(int)[mobem.index].values

        self.mobem = mobem.astype(int).T

    def get_data(self):
        return self.mobem.copy()

    def filter(self, subset=None, min_events=5):
        df = self.get_data()

        # Subset matrices
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        # Minimum number of events
        df = df[df.sum(1) >= min_events]

        return df

    @staticmethod
    def mobem_feature_to_gene(f):
        if f.endswith("_mut"):
            genes = {f.split("_")[0]}

        elif f.startswith("gain.") or f.startswith("loss."):
            genes = {
                g
                for fs in f.split("..")
                if not (fs.startswith("gain.") or fs.startswith("loss."))
                for g in fs.split(".")
                if g != ""
            }

        else:
            raise ValueError(f"{f} is not a valid MOBEM feature.")

        return genes

    @staticmethod
    def mobem_feature_type(f):
        if f.endswith("_mut"):
            return "Mutation"

        elif f.startswith("gain."):
            return "CN gain"

        elif f.startswith("loss."):
            return "CN loss"

        elif f == "msi_status":
            return f

        else:
            raise ValueError(f"{f} is not a valid MOBEM feature.")


class CopyNumber:
    def __init__(
        self,
        cnv_file="copy_number/cnv_abs_copy_number_picnic_20191101.csv.gz",
        gistic_file="copy_number/cnv_gistic_20191101.csv.gz",
        segmentation_file="copy_number/Summary_segmentation_data_994_lines_picnic.csv.gz",
        calculate_deletions=False,
        calculate_amplifications=False,
    ):
        self.ss_obj = Sample()

        self.copynumber = pd.read_csv(f"{DPATH}/{cnv_file}", index_col=0)

        self.ploidy = self.ss_obj.samplesheet["ploidy"]

        self.copynumber_seg = pd.read_csv(f"{DPATH}/{segmentation_file}")

        self.gistic = pd.read_csv(
            f"{DPATH}/{gistic_file}", index_col="gene_symbol"
        ).drop(columns=["gene_id"])

        if calculate_deletions:
            self.copynumber_del = pd.DataFrame(
                {
                    s: self.copynumber[s].apply(
                        lambda v: CopyNumber.is_deleted(v, self.ploidy[s])
                    )
                    for s in self.copynumber
                    if s in self.ploidy
                }
            )

        if calculate_amplifications:
            self.copynumber_amp = pd.DataFrame(
                {
                    s: self.copynumber[s].apply(
                        lambda v: CopyNumber.is_amplified(v, self.ploidy[s])
                    )
                    for s in self.copynumber
                    if s in self.ploidy
                }
            )

    def get_data(self, dtype="matrix"):
        if dtype == "del":
            res = self.copynumber_del.copy()

        elif dtype == "amp":
            res = self.copynumber_amp.copy()

        elif dtype == "gistic":
            res = self.gistic.copy()

        else:
            res = self.copynumber.copy()

        return res

    def filter(self, subset=None, dtype="cn"):
        df = self.get_data(dtype=dtype)

        # Subset matrices
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        return df

    def ploidy_from_segments(
        self, seg_file="copy_number/Summary_segmentation_data_994_lines_picnic.csv.gz"
    ):
        return pd.Series(
            {
                s: self.calculate_ploidy(df)[1]
                for s, df in self.copynumber_seg.groupby("model_id")
            }
        )

    @classmethod
    def genomic_instability(
        cls, seg_file="copy_number/Summary_segmentation_data_994_lines_picnic.csv.gz"
    ):
        # Import segments
        cn_seg = pd.read_csv(f"{DPATH}/{seg_file}")

        # Use only autosomal chromosomes
        cn_seg = cn_seg[~cn_seg["chr"].isin(["chrX", "chrY"])]
        cn_seg = cn_seg[~cn_seg["chr"].isin([23, 24])]

        # Segment length
        cn_seg = cn_seg.assign(length=cn_seg.eval("end - start"))

        # Calculate genomic instability
        s_instability = {}
        for s, df in cn_seg.groupby("model_id"):
            s_ploidy = np.round(cls.calculate_ploidy(df), 0)

            s_chr = []
            # c, c_df = list(df.groupby("chr"))[0]
            for c, c_df in df.groupby("chr"):
                c_gain = (
                    c_df[c_df["copy_number"] > s_ploidy]["length"].sum()
                    / c_df["length"].sum()
                )
                c_loss = (
                    c_df[c_df["copy_number"] < s_ploidy]["length"].sum()
                    / c_df["length"].sum()
                )
                s_chr.append(c_gain + c_loss)

            s_instability[s] = np.mean(s_chr)

        s_instability = pd.Series(s_instability)
        return s_instability

    @classmethod
    def calculate_ploidy(cls, cn_seg):
        # Use only autosomal chromosomes
        cn_seg = cn_seg[~cn_seg["chr"].isin(["chrX", "chrY"])]
        cn_seg = cn_seg[~cn_seg["chr"].isin([23, 24])]

        ploidy = cls.weighted_median(cn_seg["copy_number"], cn_seg["length"])

        return ploidy

        # cn_seg = cn_seg.assign(length=cn_seg["end"] - cn_seg["start"])
        # cn_seg = cn_seg.assign(
        #     cn_by_length=cn_seg["length"] * (cn_seg["copy_number"] + 1)
        # )
        #
        # chrm = (
        #     cn_seg.groupby("chr")["cn_by_length"]
        #     .sum()
        #     .divide(cn_seg.groupby("chr")["length"].sum())
        #     - 1
        # )
        #
        # ploidy = (cn_seg["cn_by_length"].sum() / cn_seg["length"].sum()) - 1
        #
        # return chrm, ploidy

    @staticmethod
    def weighted_median(data, weights):
        # Origingal code: https://gist.github.com/tinybike/d9ff1dad515b66cc0d87
        data, weights = np.array(data).squeeze(), np.array(weights).squeeze()
        s_data, s_weights = map(np.array, zip(*sorted(zip(data, weights))))
        midpoint = 0.5 * sum(s_weights)

        if any(weights > midpoint):
            w_median = (data[weights == np.max(weights)])[0]

        else:
            cs_weights = np.cumsum(s_weights)
            idx = np.where(cs_weights <= midpoint)[0][-1]
            if cs_weights[idx] == midpoint:
                w_median = np.mean(s_data[idx : idx + 2])
            else:
                w_median = s_data[idx + 1]

        return w_median

    @staticmethod
    def is_amplified(
        cn, ploidy, cn_threshold_low=5, cn_thresholds_high=9, ploidy_threshold=2.7
    ):
        if (ploidy <= ploidy_threshold) and (cn >= cn_threshold_low):
            return 1

        elif (ploidy > ploidy_threshold) and (cn >= cn_thresholds_high):
            return 1

        else:
            return 0

    @staticmethod
    def is_deleted(cn, ploidy, ploidy_threshold=2.7):
        if (ploidy <= ploidy_threshold) and (cn == 0):
            return 1

        elif (ploidy > ploidy_threshold) and (cn < (ploidy - ploidy_threshold)):
            return 1

        else:
            return 0


class Methylation:
    """
    Import module for Illumina Methylation 450k arrays
    """

    def __init__(
        self, methy_gene_promoter="methylation/methy_beta_gene_promoter.csv.gz"
    ):
        self.methy_promoter = pd.read_csv(f"{DPATH}/{methy_gene_promoter}", index_col=0)

    def get_data(self):
        return self.methy_promoter.copy()

    def filter(self, subset=None):
        df = self.get_data()

        # Subset matrices
        if subset is not None:
            df = df.loc[:, df.columns.isin(subset)]

        return df

    @staticmethod
    def discretise(beta):
        if beta < 0.33:
            return "umethylated"
        elif beta > 0.66:
            return "methylated"
        else:
            return "hemimethylated"


class CopyNumberSegmentation:
    def __init__(
        self, cnv_file="copy_number/Summary_segmentation_data_994_lines_picnic.csv.gz"
    ):
        self.copynumber = pd.read_csv(f"{DPATH}/{cnv_file}")

    def get_data(self):
        return self.copynumber.copy()

    def filter(self, subset=None):
        df = self.get_data()

        # Subset matrices
        if subset is not None:
            df = df[df["model_id"].isin(subset)]

        return df


class PipelineResults:
    def __init__(
        self,
        results_dir,
        import_fc=False,
        import_bagel=False,
        import_mageck=False,
        mageck_fdr_thres=0.1,
        fc_thres=-0.5,
    ):
        self.results_dir = results_dir

        if import_fc:
            self.fc_thres = fc_thres
            self.fc, self.fc_c, self.fc_cq, self.fc_s, self.fc_b = (
                self.import_fc_results()
            )
            LOG.info("Fold-changes imported")

        if import_bagel:
            self.bf, self.bf_q, self.bf_b = self.import_bagel_results()
            LOG.info("BAGEL imported")

        if import_mageck:
            self.mageck_fdr_thres = mageck_fdr_thres
            self.mdep_fdr, self.mdep_bin = self.import_mageck_results()
            LOG.info("MAGeCK imported")

    def import_bagel_results(self):
        # Bayesian factors
        bf = pd.read_csv(
            f"{self.results_dir}/_BayesianFactors.tsv", sep="\t", index_col=0
        )

        # Quantile normalised bayesian factors
        bf_q = pd.read_csv(
            f"{self.results_dir}/_scaledBayesianFactors.tsv", sep="\t", index_col=0
        )

        # Binarised bayesian factors
        bf_b = pd.read_csv(
            f"{self.results_dir}/_binaryDepScores.tsv", sep="\t", index_col=0
        )

        return bf, bf_q, bf_b

    def import_fc_results(self):
        # Fold-changes
        fc = pd.read_csv(f"{self.results_dir}/_logFCs.tsv", sep="\t", index_col=0)

        # Copy-number corrected fold-changes
        fc_c = pd.read_csv(
            f"{self.results_dir}/_corrected_logFCs.tsv", sep="\t", index_col=0
        )

        # Quantile normalised copy-number corrected fold-changes
        fc_cq = pd.read_csv(
            f"{self.results_dir}/_qnorm_corrected_logFCs.tsv", sep="\t", index_col=0
        )

        # Scale fold-changes
        fc_s = CRISPR.scale(fc)

        # Fold-change binary
        fc_b = (fc_s < self.fc_thres).astype(int)

        return fc, fc_c, fc_cq, fc_s, fc_b

    def import_mageck_results(self):
        # Dependencies FDR
        mdep_fdr = pd.read_csv(
            f"{self.results_dir}/_MageckFDRs.tsv", sep="\t", index_col=0
        )
        mdep_bin = (mdep_fdr < self.mageck_fdr_thres).astype(int)

        return mdep_fdr, mdep_bin



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import re
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib as mpl
import matplotlib.pyplot as plt
from crispy import CrispyPlot


class Modifications:
    AA = {
        "A": dict(abbreviation="Ala", name="Alanine"),
        "R": dict(abbreviation="Arg", name="Arginine"),
        "N": dict(abbreviation="Asn", name="Asparagine"),
        "D": dict(abbreviation="Asp", name="Aspartic acid"),
        "C": dict(abbreviation="Cys", name="Cysteine"),
        "Q": dict(abbreviation="Gln", name="Glutamine"),
        "E": dict(abbreviation="Glu", name="Glutamic acid"),
        "G": dict(abbreviation="Gly", name="Glycine"),
        "H": dict(abbreviation="His", name="Histidine"),
        "I": dict(abbreviation="Ile", name="Isoleucine"),
        "M": dict(abbreviation="Met", name="Methionine"),
        "L": dict(abbreviation="Leu", name="Leucine"),
        "K": dict(abbreviation="Lys", name="Lysine"),
        "F": dict(abbreviation="Phe", name="Phenylalanine"),
        "P": dict(abbreviation="Pro", name="Proline"),
        "S": dict(abbreviation="Ser", name="Serine"),
        "T": dict(abbreviation="Thr", name="Threonine"),
        "W": dict(abbreviation="Trp", name="Tryptophan"),
        "Y": dict(abbreviation="Tyr", name="Tyrosine"),
        "V": dict(abbreviation="Val", name="Valine"),
        "*": dict(abbreviation="STOP", name="STOP"),
    }

    @classmethod
    def get_order(cls, field="abbreviation"):
        order = cls.AA.keys()

        if field is not None:
            order = [cls.AA[i][field] for i in order]

        return order

    @classmethod
    def get_names(cls, aas=None, field="abbreviation"):
        if aas is None:
            aas = cls.AA.keys()
        return {aa: cls.AA[aa][field] for aa in aas if aa in cls.AA}


class BaseEditor:
    PAM_WT_CAS9 = "NGG"

    def __init__(
        self, edit_window=None, pam=None, base_edit=None, len_guide=None, name=None
    ):
        self.edit_window = edit_window
        self.pam = pam
        self.base_edit = base_edit
        self.len_guide = len_guide
        self.name = name

    @staticmethod
    def base_complement(nucleotide):
        if nucleotide == "C":
            return "G"
        elif nucleotide == "G":
            return "C"
        elif nucleotide == "A":
            return "T"
        elif nucleotide == "T":
            return "A"

    @staticmethod
    def parse_coordinates(genomic_coordinates):
        chrm, pos = genomic_coordinates.split(":")
        chrm, start, end = chrm, int(pos.split("-")[0]), int(pos.split("-")[1])
        return chrm, start, end

    def assert_guide(self, guide):
        assert (
            len(guide) == self.len_guide
        ), f"Guide size: expected {self.len_guide} got {len(guide)}: {guide}"

        pam_regex = self.pam.replace("N", ".")
        assert re.search(
            f"{pam_regex}$", guide
        ), f"PAM mismatch: expected {self.pam} got {guide[-len(self.pam):]}"

    def split_guide(self, guide):
        start = guide[: self.edit_window[0]]
        edit = guide[self.edit_window[0] : self.edit_window[1]]
        end = guide[self.edit_window[1] : -len(self.pam)]

        pam = guide[-len(self.pam) :]

        return start, edit, end, pam

    def print_guide(self, guide):
        self.assert_guide(guide)

        start, edit, end, pam = self.split_guide(guide)

        print(f"{start} {edit} {end} [{pam}]")

    def edit_guide(self, guide, guide_strand, target_strand):
        start, edit, end, pam = self.split_guide(guide)

        if guide_strand == target_strand:
            edit = edit.replace(
                self.base_complement(self.base_edit[0]),
                self.base_complement(self.base_edit[1]),
            )
        else:
            edit = edit.replace(self.base_edit[0], self.base_edit[1])

        return start + edit + end + pam

    def to_vep(self, guide_original, guide_edited):
        start_win, original, end_win, pam = self.split_guide(guide_original)
        _, edited, _, _ = self.split_guide(guide_edited)

        if original == edited:
            return None

        else:
            idx_start = min([i for i, bp in enumerate(original) if bp != edited[i]])
            idx_end = len(original) - min(
                [i for i, bp in enumerate(original[::-1]) if bp != edited[::-1][i]]
            )

            edit = f"{original[idx_start:idx_end]}/{edited[idx_start:idx_end]}"

            return edit, len(start_win) + idx_start, len(start_win) + (idx_end - 1)

    def list_base_edits(self, guide_original, guide_edited):
        start_win, original, end_win, pam = self.split_guide(guide_original)
        _, edited, _, _ = self.split_guide(guide_edited)

        if original == edited:
            return None

        else:
            return [
                (f"{original[i]}/{edited[i]}", len(start_win) + i)
                for i, bp in enumerate(original)
                if original[i] != edited[i]
            ]


class CytidineBaseEditor(BaseEditor):
    def __init__(
        self,
        edit_window=(3, 8),
        pam=BaseEditor.PAM_WT_CAS9,
        len_guide=23,
        name="Cytidine base editor",
    ):
        super().__init__(
            edit_window=edit_window,
            pam=pam,
            base_edit=("C", "T"),
            len_guide=len_guide,
            name=name,
        )


class AdenineBaseEditor(BaseEditor):
    def __init__(
        self,
        edit_window=(3, 8),
        pam=BaseEditor.PAM_WT_CAS9,
        len_guide=23,
        name="Adenine base editor",
    ):
        super().__init__(
            edit_window=edit_window,
            pam=pam,
            base_edit=("A", "G"),
            len_guide=len_guide,
            name=name,
        )


class BeditPlot(CrispyPlot):
    # TODO: Test situtation where not all 20 aa and stop codon are present

    @classmethod
    def region_scatter(cls, dataframe, y_var="aa_pos", x_var="score"):
        plot_df = dataframe.dropna(subset=[y_var, x_var])
        plot_df["z_var"] = cls.density_interpolate(plot_df[y_var], plot_df[x_var])

        fig, ax = plt.subplots(
            1, 1, figsize=(min(len(set(plot_df[y_var])) * 0.01, 15), 2.0), dpi=600
        )

        ax.scatter(
            plot_df[y_var],
            plot_df[x_var],
            c=plot_df["z_var"],
            marker="o",
            edgecolor="",
            cmap="Spectral_r",
            s=2,
        )

        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)

        ax.set_xlabel("Position")
        ax.set_ylabel("Functional score")

        return ax

    @classmethod
    def aa_grid(
        cls, dataframe=None, aggfunc=len, index_var="mutant", col_var="wildtype"
    ):
        order = Modifications.get_order(field="abbreviation")

        plot_df = pd.pivot_table(
            dataframe[[index_var, col_var]],
            index=index_var,
            columns=col_var,
            aggfunc=aggfunc,
            fill_value=np.nan,
        ).loc[order, order]

        #
        f, axs = plt.subplots(
            2,
            2,
            sharex="col",
            sharey="row",
            gridspec_kw=dict(height_ratios=[1, 4], width_ratios=[4, 1]),
        )

        g = sns.heatmap(
            plot_df, cmap="Greys", annot=True, cbar=False, linewidths=0.5, ax=axs[1, 0]
        )

        axs[1, 0].set_xlabel("Wild-type")
        axs[1, 0].set_ylabel("Mutant")
        axs[1, 0].grid(lw=0.5, ls="--", color=cls.PAL_DBGD[2])
        sns.despine(right=False, top=False, ax=axs[1, 0])

        for i in [0, 1]:
            i_sums = plot_df.sum(i)

            norm = mpl.colors.Normalize(vmin=0, vmax=i_sums.max())
            colors = [plt.cm.Greys(norm(v)) for v in i_sums]

            barfun = axs[i, i].barh if i else axs[i, i].bar
            barfun(np.arange(len(order)) + 0.5, i_sums, color=colors)

            axs[i, i].set_xlabel("Count" if i else "")
            axs[i, i].set_ylabel("" if i else "Count")
            sns.despine(ax=axs[i, i])

        axs[0, 1].axis("off")

        plt.subplots_adjust(hspace=0, wspace=0)

        return g

    @classmethod
    def aa_countplot(cls, dataframe, index_var="mutant", col_var="wildtype"):
        plot_df = pd.concat(
            [
                dataframe[index_var].value_counts().rename("Mutant"),
                dataframe[col_var].value_counts().rename("Wild-type"),
            ],
            axis=1,
            sort=False,
        )

        plot_df = plot_df.unstack().reset_index()
        plot_df.columns = ["type", "aa", "count"]

        order = Modifications.get_order()

        pal = {"Mutant": cls.PAL_DBGD[0], "Wild-type": cls.PAL_DBGD[2]}

        sns.barplot(
            "count", "aa", "type", data=plot_df, palette=pal, order=order, orient="h"
        )

        plt.grid(lw=0.3, ls="-", color=cls.PAL_DBGD[2], axis="x", zorder=0)

        plt.legend(frameon=False, prop=dict(size=4))

        plt.xlabel("AA count")
        plt.ylabel("")



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import random
import logging
import operator
import numpy as np
import pandas as pd
import pkg_resources
from crispy import CrispyPlot
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from scipy.stats.distributions import hypergeom
from statsmodels.stats.multitest import multipletests


dpath = pkg_resources.resource_filename("crispy", "data/")


class SSGSEA(object):
    @classmethod
    def gsea(cls, dataset, signature, permutations=0):
        # Sort data-set by values
        _dataset = list(
            zip(*sorted(dataset.items(), key=operator.itemgetter(1), reverse=True))
        )
        genes, expression = _dataset[0], _dataset[1]

        # Signature overlapping the data-set
        _signature = set(signature).intersection(genes)

        # Check signature overlap
        e_score, p_value = np.NaN, np.NaN
        hits, running_hit = [], []
        if len(_signature) != 0:

            # ---- Calculate signature enrichment score
            n, sig_size = len(genes), len(_signature)
            nh = n - sig_size
            nr = sum([abs(dataset[g]) for g in _signature])

            e_score = cls.__es(
                genes, expression, _signature, nr, nh, n, hits, running_hit
            )

            # ---- Calculate statistical enrichment
            # Generate random signatures sampled from the data-set genes
            if permutations > 0:
                count = 0

                for i in range(permutations):
                    r_signature = random.sample(genes, sig_size)

                    r_nr = sum([abs(dataset[g]) for g in r_signature])

                    r_es = cls.__es(genes, expression, r_signature, r_nr, nh, n)

                    if (r_es >= e_score >= 0) or (r_es <= e_score < 0):
                        count += 1

                # If no permutation was above the Enrichment score the p-value is lower than 1 divided by the number
                # of permutations
                p_value = 1 / permutations if count == 0 else count / permutations

            else:
                p_value = np.nan

        return e_score, p_value, hits, running_hit

    @staticmethod
    def __es(genes, expression, signature, nr, nh, n, hits=None, running_hit=None):
        hit, miss, es, r = 0, 0, 0, 0
        for i in range(n):
            if genes[i] in signature:
                hit += abs(expression[i]) / nr

                if hits is not None:
                    hits.append(1)

            else:
                miss += 1 / nh

                if hits is not None:
                    hits.append(0)

            r = hit - miss

            if running_hit is not None:
                running_hit.append(r)

            if abs(r) > abs(es):
                es = r

        return es


class Enrichment:
    """
    Gene enrichment analysis class.

    """

    def __init__(
        self, gmts, sig_min_len=5, verbose=0, padj_method="fdr_bh", permutations=0
    ):

        self.verbose = verbose
        self.padj_method = padj_method
        self.permutations = permutations

        self.sig_min_len = sig_min_len

        self.gmts = {f: self.read_gmt(f"{dpath}/pathways/{f}") for f in gmts}

    def __assert_gmt_file(self, gmt_file):
        assert gmt_file in self.gmts, f"{gmt_file} not in gmt files: {self.gmts.keys()}"

    def __assert_signature(self, gmt_file, signature):
        self.__assert_gmt_file(gmt_file)
        assert signature in self.gmts[gmt_file], f"{signature} not in {gmt_file}"

    @staticmethod
    def read_gmt(file_path, subset=None, min_size=15):
        with open(file_path) as f:
            signatures = {}

            for l in f.readlines():
                sig_name = l.split("\t")[0]

                sig_genes = set(l.strip().split("\t")[2:])

                if subset is not None:
                    sig_genes = sig_genes.intersection(subset)

                if len(sig_genes) > min_size:
                    signatures[sig_name] = sig_genes

        return signatures

    @classmethod
    def read_signature(cls, gmt_file, signature):
        return cls.read_gmt(gmt_file)[signature]

    @classmethod
    def signature(cls, signature):
        if signature.startswith("HALLMARK_"):
            gs_file = "h.all.v7.1.symbols.gmt"

        elif signature.startswith("KEGG_"):
            gs_file = "c2.cp.kegg.v7.1.symbols.gmt"

        elif signature.startswith("REACTOME_"):
            gs_file = "c2.cp.v7.1.symbols.gmt"

        elif signature.startswith("SIG_"):
            gs_file = "c2.cp.v7.1.symbols.gmt"

        elif signature.startswith("CSR_"):
            gs_file = "c6.all.v7.1.symbols.gmt"

        else:
            gs_file = "c5.all.v7.1.symbols.gmt"

        return cls.read_gmt(f"{dpath}/pathways/{gs_file}")[signature]

    def gsea(self, values, signature):
        return SSGSEA.gsea(values.to_dict(), signature, permutations=self.permutations)

    def gsea_enrichments(self, values):
        return pd.concat(
            [self.gsea_enrichment(values, gmt_file=g).assign(gmt=g) for g in self.gmts]
        ).sort_values("e_score")

    def gsea_enrichment(self, values, gmt_file):
        self.__assert_gmt_file(gmt_file)

        geneset = self.gmts[gmt_file]

        if self.verbose > 0 and type(values) == pd.Series:
            logging.getLogger("DTrace").info(f"Values={values.name}")

        ssgsea = []
        for gset in geneset:
            if self.verbose > 1:
                logging.getLogger("DTrace").info(f"Gene-set={gset}")

            gset_len = len({i for i in geneset[gset] if i in values.index})

            e_score, p_value, _, _ = self.gsea(values, geneset[gset])

            ssgsea.append(
                dict(gset=gset, e_score=e_score, p_value=p_value, len=gset_len)
            )

        ssgsea = pd.DataFrame(ssgsea).set_index("gset").sort_values("e_score")

        if self.sig_min_len is not None:
            ssgsea = ssgsea.query(f"len >= {self.sig_min_len}")

        if self.permutations > 0:
            ssgsea["adj.p_value"] = multipletests(
                ssgsea["p_value"], method=self.padj_method
            )[1]

        return ssgsea

    def get_signature(self, gmt_file, signature):
        self.__assert_signature(gmt_file, signature)
        return self.gmts[gmt_file][signature]

    def plot(self, values, gmt_file, signature, vertical_lines=False, shade=False):
        if type(signature) == str:
            gset = self.get_signature(gmt_file, signature)

        e_score, p_value, hits, running_hit = self.gsea(values, gset)

        ax = GSEAplot.plot_gsea(
            hits,
            running_hit,
            dataset=values.to_dict(),
            vertical_lines=vertical_lines,
            shade=shade,
        )

        ax[0].set_title(f"{signature}: NES={e_score:.2f}; p-val={p_value:.2g}")

        return ax

    @staticmethod
    def hypergeom_test(signature, background, sublist):
        """
        Performs hypergeometric test

        Arguements:
                signature: {string} - Signature IDs
                background: {string} - Background IDs
                sublist: {string} - Sub-set IDs

        # hypergeom.sf(x, M, n, N, loc=0)
        # M: total number of objects,
        # n: total number of type I objects
        # N: total number of type I objects drawn without replacement

        """
        pvalue = hypergeom.sf(
            len(sublist.intersection(signature)),
            len(background),
            len(background.intersection(signature)),
            len(sublist),
        )

        intersection = len(sublist.intersection(signature))

        return pvalue, intersection

    def hypergeom_enrichments(self, sublist, background, gmt_file):
        self.__assert_gmt_file(gmt_file)

        geneset = self.gmts[gmt_file]

        ssgsea_geneset = []
        for gset in geneset:
            if self.verbose > 0:
                logging.getLogger("DTrace").info(f"Gene-set={gset}")

            p_value, intersection = self.hypergeom_test(
                signature=geneset[gset], background=background, sublist=sublist
            )

            if intersection >= self.sig_min_len:
                ssgsea_geneset.append(
                    dict(
                        gset=gset,
                        p_value=p_value,
                        len_sig=len(geneset[gset]),
                        len_intersection=intersection,
                    )
                )

        ssgsea_geneset = (
            pd.DataFrame(ssgsea_geneset).set_index("gset").sort_values("p_value")
        )
        ssgsea_geneset["adj.p_value"] = multipletests(
            ssgsea_geneset["p_value"], method=self.padj_method
        )[1]

        return ssgsea_geneset

    @staticmethod
    def one_sided_pvalue(escore, escores):
        count = np.sum((escores >= escore) if escore >= 0 else (escores <= escore))

        p_value = 1 / len(escores) if count == 0 else count / len(escores)

        return p_value


class GSEAplot(CrispyPlot):
    @classmethod
    def plot_gsea(
        cls, hits, running_hit, dataset=None, vertical_lines=False, shade=False
    ):
        x, y = np.array(range(len(hits))), np.array(running_hit)

        if dataset is not None:
            gs = GridSpec(2, 1, height_ratios=[3, 2], hspace=0.1)
            axs = [plt.subplot(gs[0]), plt.subplot(gs[1])]

        else:
            axs = [plt.gca()]

        # GSEA running hit
        axs[0].plot(x, y, "-", c=cls.PAL_DBGD[0])

        if shade:
            axs[0].fill_between(x, 0, y, alpha=0.5, color=cls.PAL_DBGD[2])

        if vertical_lines:
            for i in x[np.array(hits, dtype="bool")]:
                axs[0].axvline(i, c=cls.PAL_DBGD[0], lw=0.3, alpha=0.2, zorder=0)

        axs[0].axhline(0, c=cls.PAL_DBGD[0], lw=0.1, ls="-")
        axs[0].set_ylabel("Enrichment score")
        axs[0].get_xaxis().set_visible(False)
        axs[0].set_xlim([0, len(x)])

        if dataset is not None:
            dataset = list(
                zip(*sorted(dataset.items(), key=operator.itemgetter(1), reverse=False))
            )

            # Data
            axs[1].scatter(x, dataset[1], c=cls.PAL_DBGD[0], linewidths=0, s=2)

            if vertical_lines:
                for i in x[np.array(hits, dtype="bool")]:
                    axs[1].axvline(i, c=cls.PAL_DBGD[0], lw=0.3, alpha=0.2, zorder=0)

            axs[1].axhline(0, c="black", lw=0.3, ls="-")
            axs[1].set_ylabel("Data value")
            axs[1].get_xaxis().set_visible(False)
            axs[1].set_xlim([0, len(x)])

        return axs[0] if dataset is None else axs



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import numpy as np
import pandas as pd
import seaborn as sns
import scipy.stats as st
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
from natsort import natsorted
from crispy.Utils import Utils
from matplotlib.patches import Arc
from collections import OrderedDict
from sklearn.metrics.ranking import auc
from crispy.CrispyPlot import CrispyPlot
from sklearn.metrics import (
    roc_auc_score,
    roc_curve,
    precision_recall_curve,
    average_precision_score,
)


class QCplot(CrispyPlot):
    @staticmethod
    def aroc_threshold(
        values, true_set=None, false_set=None, fpr_thres=0.01, return_curve=False
    ):
        if true_set is None:
            true_set = Utils.get_essential_genes(return_series=False)

        if false_set is None:
            false_set = Utils.get_non_essential_genes(return_series=False)

        index_set = true_set.union(false_set)

        rank = values[values.index.isin(index_set)]
        y_true = rank.index.isin(true_set).astype(int)

        fpr, tpr, thres = roc_curve(y_true, -rank)

        auc_fpr = roc_auc_score(y_true, -rank, max_fpr=fpr_thres)

        if fpr_thres is not None:
            fc_thres_fpr = -min(thres[fpr <= fpr_thres])

        else:
            fc_thres_fpr = None

        res = (
            (auc_fpr, fc_thres_fpr, fpr, tpr)
            if return_curve
            else (auc_fpr, fc_thres_fpr)
        )

        return res

    @staticmethod
    def precision_recall_curve(
        values, true_set=None, false_set=None, fdr_thres=0.01, return_curve=False
    ):
        if true_set is None:
            true_set = Utils.get_essential_genes(return_series=False)

        if false_set is None:
            false_set = Utils.get_non_essential_genes(return_series=False)

        index_set = true_set.union(false_set)

        rank = values[values.index.isin(index_set)]
        y_true = rank.index.isin(true_set).astype(int)

        ap = average_precision_score(y_true, -rank)

        precision, recall, thres = precision_recall_curve(y_true, -rank)
        recall_fdr = recall[precision > (1 - fdr_thres)].max()

        res = (
            (ap, recall_fdr, precision, recall, thres)
            if return_curve
            else (ap, recall_fdr)
        )

        return res

    @staticmethod
    def recall_curve(rank, index_set=None, min_events=None):
        """
        Calculate x and y of recall curve.

        :param rank: pandas.Series

        :param index_set: pandas.Series
            indices in rank

        :param min_events: int or None, optional
            Number of minimum number of index_set to calculate curve

        :return:
        """
        x = rank.sort_values().dropna()

        # Observed cumsum
        if index_set is None:
            index_set = Utils.get_essential_genes(return_series=False)

        y = x.index.isin(index_set)

        if (min_events is not None) and (sum(y) < min_events):
            return None

        y = np.cumsum(y) / sum(y)

        # Rank fold-changes
        x = st.rankdata(x) / x.shape[0]

        # Calculate AUC
        xy_auc = auc(x, y)

        return x, y, xy_auc

    @staticmethod
    def pr_curve(rank, true_set=None, false_set=None, min_events=10):
        if true_set is None:
            true_set = Utils.get_essential_genes(return_series=False)

        if false_set is None:
            false_set = Utils.get_non_essential_genes(return_series=False)

        index_set = true_set.union(false_set)

        rank = rank[rank.index.isin(index_set)]

        if len(rank) == 0:
            return np.nan

        y_true = rank.index.isin(true_set).astype(int)

        if sum(y_true) < min_events:
            return np.nan

        return roc_auc_score(y_true, -rank)

    @classmethod
    def recall_curve_discretise(cls, rank, discrete, thres, min_events=10):
        discrete = discrete.apply(Utils.bin_cnv, args=(thres,))

        aucs = pd.DataFrame(
            [
                {discrete.name: g, "auc": cls.recall_curve(rank, df.index)[2]}
                for g, df in discrete.groupby(discrete)
                if df.shape[0] >= min_events
            ]
        )

        return aucs

    @classmethod
    def bias_boxplot(
        cls,
        data,
        x="copy_number",
        y="auc",
        hue=None,
        despine=False,
        notch=True,
        add_n=False,
        n_text_y=None,
        n_text_offset=0.05,
        palette=None,
        ax=None,
        tick_base=0.1,
        order=None,
        hue_order=None,
        n_fontsize=3.5,
        draw_violin=False,
    ):
        if ax is None:
            ax = plt.gca()

        order = natsorted(set(data[x])) if order is None else order

        if palette is None and hue is not None:
            hue_order = natsorted(set(data[hue])) if hue_order is None else hue_order
            palette = dict(
                zip(*(hue_order, cls.get_palette_continuous(len(hue_order))))
            )

        elif palette is None:
            palette = dict(zip(*(order, cls.get_palette_continuous(len(order)))))

        if draw_violin:
            sns.violinplot(
                x,
                y,
                data=data,
                hue=hue,
                order=order,
                palette=palette,
                saturation=1.0,
                hue_order=hue_order,
                inner="quartile",
                linewidth=.5,
                ax=ax,
            )
        else:
            sns.boxplot(
                x,
                y,
                data=data,
                hue=hue,
                notch=notch,
                order=order,
                palette=palette,
                saturation=1.0,
                showcaps=False,
                hue_order=hue_order,
                medianprops=cls.MEDIANPROPS,
                flierprops=cls.FLIERPROPS,
                whiskerprops=cls.WHISKERPROPS,
                boxprops=cls.BOXPROPS,
                ax=ax,
            )

        if despine:
            sns.despine(ax=ax)

        if add_n:
            text_y = min(data[y]) - n_text_offset if n_text_y is None else n_text_y

            for i, c in enumerate(order):
                n = np.sum(data[x] == c)
                ax.text(
                    i, text_y, f"N={n}", ha="center", va="bottom", fontsize=n_fontsize
                )

            y_lim = ax.get_ylim()
            ax.set_ylim(text_y, y_lim[1])

        if tick_base is not None:
            ax.yaxis.set_major_locator(plticker.MultipleLocator(base=tick_base))

        return ax

    @classmethod
    def plot_cumsum_auc(
        cls,
        df,
        index_set,
        ax=None,
        palette=None,
        legend=True,
        plot_mean=False,
        legend_prop=None,
        color="#484848",
    ):
        """
        Plot cumulative sum of values X considering index_set list.

        :param DataFrame df: measurements values
        :param Series index_set: list of indexes in X
        :param ax: matplotlib Axes
        :param dict palette: palette
        :param Bool legend: Plot legend
        :param Bool plot_mean: Plot curve with the mean across all columns in X
        :param legend_prop:
        :param color:

        :return matplotlib Axes, plot_stats dict:
        """

        if ax is None:
            ax = plt.gca()

        if type(index_set) == set:
            index_set = pd.Series(list(index_set)).rename("geneset")

        plot_stats = dict(auc={})

        # Calculate recall curves
        xy_curves = {f: cls.recall_curve(df[f], index_set) for f in list(df)}

        # Build curve
        for f in xy_curves:
            x, y, xy_auc = xy_curves[f]

            # Calculate AUC
            plot_stats["auc"][f] = xy_auc

            # Plot
            c = color if palette is None else palette[f]
            ax.plot(
                x,
                y,
                label="%s: %.2f" % (f, xy_auc) if (legend is True) else None,
                lw=1.0,
                c=c,
                alpha=0.8,
            )

        # Mean
        if plot_mean:
            x, y, xy_auc = cls.recall_curve(df.mean(1), index_set)

            plot_stats["auc"]["mean"] = xy_auc

            ax.plot(
                x, y, label=f"Mean: {xy_auc:.2f}", ls="--", c="red", lw=2.5, alpha=0.8
            )

        # Random
        ax.plot((0, 1), (0, 1), "k--", lw=0.3, alpha=0.5)

        # Limits
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)

        # Labels
        ax.set_xlabel("Ranked genes (fold-change)")
        ax.set_ylabel(f"Recall {index_set.name}")

        if legend is True:
            ax.legend(loc=4, frameon=False, prop=legend_prop)

        return ax, plot_stats

    @classmethod
    def aucs_scatter(
        cls, x, y, data, diagonal_line=True, rugplot=False, s=4, marker="x", lw=0.05
    ):
        ax = plt.gca()

        ax.scatter(
            data[x],
            data[y],
            c=cls.PAL_DBGD[0],
            s=s,
            marker=marker,
            lw=lw,
            edgecolor="white",
        )

        if rugplot:
            sns.rugplot(
                data[x], height=0.02, axis="x", c=cls.PAL_DBGD[0], lw=0.3, ax=ax
            )
            sns.rugplot(
                data[y], height=0.02, axis="y", c=cls.PAL_DBGD[0], lw=0.3, ax=ax
            )

        if diagonal_line:
            (x0, x1), (y0, y1) = ax.get_xlim(), ax.get_ylim()
            lims = [max(x0, y0), min(x1, y1)]
            ax.plot(lims, lims, "k-", lw=0.3, zorder=0)

        ax.set_xlabel("{} fold-changes AURCs".format(x.capitalize()))
        ax.set_ylabel("{} fold-changes AURCs".format(y.capitalize()))

        return ax

    @classmethod
    def aucs_scatter_pairgrid(cls, data, set_lim=True, axhlines=True, axvlines=True):
        g = sns.PairGrid(data, despine=False)

        g.map_offdiag(
            plt.scatter, color=cls.PAL_DBGD[0], s=4, marker="o", lw=0.5, alpha=0.7
        )

        g.map_diag(plt.hist, color=cls.PAL_DBGD[0])

        if set_lim:
            g.set(ylim=(0, 1), xlim=(0, 1))

        for ax_indices in [
            np.tril_indices_from(g.axes, -1),
            np.triu_indices_from(g.axes, 1),
        ]:
            for i, j in zip(*ax_indices):
                if axhlines:
                    g.axes[i, j].axhline(0.5, ls=":", lw=0.3, zorder=0, color="black")

                if axvlines:
                    g.axes[i, j].axvline(0.5, ls=":", lw=0.3, zorder=0, color="black")

        for ax_indices in [
            np.tril_indices_from(g.axes, -1),
            np.triu_indices_from(g.axes, 1),
        ]:
            for i, j in zip(*ax_indices):
                (x0, x1), (y0, y1) = g.axes[i, j].get_xlim(), g.axes[i, j].get_ylim()
                lims = [max(x0, y0), min(x1, y1)]

                g.axes[i, j].plot(lims, lims, "k-", lw=0.3, zorder=0)

        return g

    @classmethod
    def plot_chromosome(
        cls,
        crispy_bed,
        ascat_bed,
        chrm,
        y_var="fold_change",
        highlight=None,
        ax=None,
        legend=False,
        scale=1e6,
        tick_base=1,
        legend_size=5,
    ):

        if ax is None:
            ax = plt.gca()

        # - Build data-frames
        # ASCAT
        ascat_ = ascat_bed.query(f"Chr == '{chrm}'")

        # CRISPR
        crispr_ = crispy_bed[crispy_bed["Chr"] == chrm]
        crispr_ = crispr_.assign(location=crispr_[["sgRNA_Start", "sgRNA_End"]].mean(1))

        crispr_gene_ = crispr_.groupby("gene")[[y_var, "location"]].mean()

        # Plot original values
        ax.scatter(
            crispr_["location"] / scale,
            crispr_[y_var],
            s=6,
            marker=".",
            lw=0,
            c=cls.PAL_DBGD[1],
            alpha=0.4,
            label="CRISPR-Cas9",
        )

        # Segment mean
        for (s, e), gp_mean in crispr_.groupby(["Start", "End"])[y_var]:
            ax.plot(
                (s / scale, e / scale),
                (gp_mean.mean(), gp_mean.mean()),
                alpha=1.0,
                c=cls.PAL_DBGD[2],
                zorder=3,
                label="CRISPR-Cas9 segment mean",
                lw=2,
            )

        # Plot segments
        for s, e, cn in ascat_[["Start", "End", "copy_number"]].values:
            ax.plot(
                (s / scale, e / scale),
                (cn, cn),
                alpha=1.0,
                c=cls.PAL_DBGD[0],
                zorder=3,
                label="Copy-number segment",
                lw=2,
            )

        # Highlight
        if highlight is not None:
            for ic, i in zip(
                *(sns.color_palette("tab20", n_colors=len(highlight)), highlight)
            ):
                if i in crispr_gene_.index:
                    ax.scatter(
                        crispr_gene_["location"].loc[i] / scale,
                        crispr_gene_[y_var].loc[i],
                        s=14,
                        marker="X",
                        lw=0,
                        c=ic,
                        alpha=0.9,
                        label=i,
                    )
        # Misc
        ax.axhline(0, lw=0.3, ls="-", color="black")

        # Cytobads
        cytobands = Utils.get_cytobands(chrm=chrm)

        for i, (s, e, t) in enumerate(cytobands[["Start", "End", "band"]].values):
            if t == "acen":
                ax.axvline(s / scale, lw=0.2, ls="-", color=cls.PAL_DBGD[0], alpha=0.1)
                ax.axvline(e / scale, lw=0.2, ls="-", color=cls.PAL_DBGD[0], alpha=0.1)

            elif not i % 2:
                ax.axvspan(s / scale, e / scale, alpha=0.1, facecolor=cls.PAL_DBGD[0])

        # Legend
        if legend:
            handles, labels = plt.gca().get_legend_handles_labels()
            by_label = OrderedDict(zip(labels, handles))
            plt.legend()
            ax.legend(
                by_label.values(),
                by_label.keys(),
                loc="center left",
                bbox_to_anchor=(1, 0.5),
                prop={"size": legend_size},
                frameon=False,
            )

        ax.set_xlim(crispr_["Start"].min() / scale, crispr_["End"].max() / scale)

        ax.tick_params(axis="both", which="major", labelsize=5)

        ax.yaxis.set_major_locator(plticker.MultipleLocator(base=tick_base))

        return ax

    @classmethod
    def plot_rearrangements(
        cls,
        brass_bedpe,
        ascat_bed,
        crispy_bed,
        chrm,
        chrm_size=None,
        xlim=None,
        scale=1e6,
        show_legend=True,
        unfold_inversions=False,
        sv_alpha=1.0,
        sv_lw=0.3,
        highlight=None,
        mark_essential=False,
    ):
        # - Define default params
        chrm_size = Utils.CHR_SIZES_HG19 if chrm_size is None else chrm_size

        xlim = (0, chrm_size[chrm]) if xlim is None else xlim

        # - Build data-frames
        # BRASS
        brass_ = brass_bedpe[
            (brass_bedpe["chr1"] == chrm) | (brass_bedpe["chr2"] == chrm)
        ]

        # ASCAT
        ascat_ = ascat_bed.query(f"chr == '{chrm}'")

        # CRISPR
        crispr_ = crispy_bed[crispy_bed["chr"] == chrm]
        crispr_ = crispr_.assign(location=crispr_[["sgrna_start", "sgrna_end"]].mean(1))

        crispr_gene_ = crispr_.groupby("gene")[["fold_change", "location"]].mean()

        if brass_.shape[0] == 0:
            return None, None, None

        # - Plot
        f, (ax1, ax2, ax3) = plt.subplots(
            3, 1, sharex="all", gridspec_kw={"height_ratios": [1, 2, 2]}
        )

        # Top panel
        ax1.axhline(0.0, lw=0.3, color=cls.PAL_DBGD[0])
        ax1.set_ylim(-1, 1)

        # Middle panel
        for i, (_, s, e, cn) in ascat_.iterrows():
            ax2.plot(
                (s / scale, e / scale),
                (cn, cn),
                alpha=1.0,
                c=cls.PAL_DBGD[2],
                zorder=3,
                label="ASCAT",
                lw=2,
            )

        # Bottom panel
        ax3.scatter(
            crispr_["location"] / scale,
            crispr_["fold_change"],
            s=1,
            alpha=0.5,
            lw=0,
            c=cls.PAL_DBGD[1],
            label="CRISPR-Cas9",
            zorder=1,
        )
        ax3.axhline(0.0, lw=0.3, color=cls.PAL_DBGD[0])

        for (s, e), gp_mean in crispr_.groupby(["start", "end"])["fold_change"]:
            ax3.plot(
                (s / scale, e / scale),
                (gp_mean.mean(), gp_mean.mean()),
                alpha=1.0,
                c=cls.PAL_DBGD[2],
                zorder=3,
                label="Segment mean",
                lw=2,
            )

        if mark_essential:
            ess = Utils.get_adam_core_essential()
            ax3.scatter(
                crispr_gene_.reindex(ess)["location"] / scale,
                crispr_gene_.reindex(ess)["fold_change"],
                s=5,
                marker="x",
                lw=0.3,
                c=cls.PAL_DBGD[1],
                alpha=0.4,
                edgecolors="#fc8d62",
                label="Core-essential",
            )

        # Highlight
        if highlight is not None:
            for ic, i in zip(
                *(sns.color_palette("tab20", n_colors=len(highlight)), highlight)
            ):
                if i in crispr_.index:
                    ax3.scatter(
                        crispr_.loc[i, "location"] / scale,
                        crispr_.loc[i]["fold_change"],
                        s=14,
                        marker="X",
                        lw=0,
                        c=ic,
                        alpha=0.9,
                        label=i,
                    )

        #
        for c1, s1, e1, c2, s2, e2, st1, st2, sv in brass_[
            [
                "chr1",
                "start1",
                "end1",
                "chr2",
                "start2",
                "end2",
                "strand1",
                "strand2",
                "svclass",
            ]
        ].values:
            stype = Utils.svtype(st1, st2, sv, unfold_inversions)
            stype_col = cls.SV_PALETTE[stype]

            zorder = 2 if stype == "tandem-duplication" else 1

            x1_mean, x2_mean = np.mean([s1, e1]), np.mean([s2, e2])

            # Plot arc
            if c1 == c2:
                angle = 0 if stype in ["tandem-duplication", "deletion"] else 180

                xy = (np.mean([x1_mean, x2_mean]) / scale, 0)

                ax1.add_patch(
                    Arc(
                        xy,
                        (x2_mean - x1_mean) / scale,
                        1.0,
                        angle=angle,
                        theta1=0,
                        theta2=180,
                        edgecolor=stype_col,
                        lw=sv_lw,
                        zorder=zorder,
                        alpha=sv_alpha,
                    )
                )

            # Plot segments
            for ymin, ymax, ax in [(-1, 0.5, ax1), (-1, 1, ax2), (0, 1, ax3)]:
                if (c1 == chrm) and (xlim[0] <= x1_mean <= xlim[1]):
                    ax.axvline(
                        x=x1_mean / scale,
                        ymin=ymin,
                        ymax=ymax,
                        c=stype_col,
                        linewidth=sv_lw,
                        zorder=zorder,
                        clip_on=False,
                        label=stype,
                        alpha=sv_alpha,
                    )

                if (c2 == chrm) and (xlim[0] <= x2_mean <= xlim[1]):
                    ax.axvline(
                        x=x2_mean / scale,
                        ymin=ymin,
                        ymax=ymax,
                        c=stype_col,
                        linewidth=sv_lw,
                        zorder=zorder,
                        clip_on=False,
                        label=stype,
                        alpha=sv_alpha,
                    )

            # Translocation label
            if stype == "translocation":
                if (c1 == chrm) and (xlim[0] <= x1_mean <= xlim[1]):
                    ax1.text(
                        x1_mean / scale,
                        0,
                        " to {}".format(c2),
                        color=stype_col,
                        ha="center",
                        fontsize=5,
                        rotation=90,
                        va="bottom",
                    )

                if (c2 == chrm) and (xlim[0] <= x2_mean <= xlim[1]):
                    ax1.text(
                        x2_mean / scale,
                        0,
                        " to {}".format(c1),
                        color=stype_col,
                        ha="center",
                        fontsize=5,
                        rotation=90,
                        va="bottom",
                    )

        #
        if show_legend:
            by_label = {
                l.capitalize(): p
                for p, l in zip(*(ax2.get_legend_handles_labels()))
                if l in cls.SV_PALETTE
            }
            ax1.legend(
                by_label.values(),
                by_label.keys(),
                loc="center left",
                bbox_to_anchor=(1.02, 0.5),
                prop={"size": 6},
                frameon=False,
            )

            by_label = {
                l: p
                for p, l in zip(*(ax2.get_legend_handles_labels()))
                if l not in cls.SV_PALETTE
            }
            ax2.legend(
                by_label.values(),
                by_label.keys(),
                loc="center left",
                bbox_to_anchor=(1.02, 0.5),
                prop={"size": 6},
                frameon=False,
            )

            by_label = {
                l: p
                for p, l in zip(*(ax3.get_legend_handles_labels()))
                if l not in cls.SV_PALETTE
            }
            ax3.legend(
                by_label.values(),
                by_label.keys(),
                loc="center left",
                bbox_to_anchor=(1.02, 0.5),
                prop={"size": 6},
                frameon=False,
            )

        #
        ax1.axis("off")

        #
        ax2.set_ylim(0, np.ceil(ascat_["copy_number"].quantile(0.9999) + 0.5))

        #
        ax2.yaxis.set_major_locator(plticker.MultipleLocator(base=2.0))
        ax3.yaxis.set_major_locator(plticker.MultipleLocator(base=1.0))

        #
        ax2.tick_params(axis="both", which="major", labelsize=6)
        ax3.tick_params(axis="both", which="major", labelsize=6)

        #
        ax1.set_ylabel("SV")
        ax2.set_ylabel("Copy-number", fontsize=7)
        ax3.set_ylabel("Loss of fitness", fontsize=7)

        #
        plt.xlabel("Position on chromosome {} (Mb)".format(chrm.replace("chr", "")))

        #
        plt.xlim(xlim[0] / scale, xlim[1] / scale)

        return ax1, ax2, ax3



#!/usr/bin/env python
# Copyright (C) 2019 Emanuel Goncalves

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from math import sqrt
from natsort import natsorted
from crispy.Utils import Utils
from adjustText import adjust_text
from matplotlib.lines import Line2D
from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import MinMaxScaler
from crispy.CrispyPlot import CrispyPlot, MidpointNormalize


class GIPlot(CrispyPlot):
    MARKERS = ["o", "X", "v", "^"]

    @classmethod
    def gi_regression_no_marginals(
        cls,
        x_gene,
        y_gene,
        plot_df,
        alpha=1.0,
        hue=None,
        style=None,
        lowess=False,
        palette=None,
        plot_reg=True,
        figsize=(3, 3),
        plot_style_legend=True,
        plot_hue_legend=True,
        ax=None,
    ):
        pal = cls.PAL_DTRACE if palette is None else palette

        plot_df = plot_df.dropna(subset=[x_gene, y_gene])

        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=figsize, dpi=600)

        markers_handles = dict()

        # Joint
        for t, df in [("none", plot_df)] if hue is None else plot_df.groupby(hue):
            for i, (n, df_style) in enumerate(
                [("none", df)] if style is None else df.groupby(style)
            ):
                ax.scatter(
                    x=df_style[x_gene],
                    y=df_style[y_gene],
                    edgecolor="w",
                    lw=0.1,
                    s=3,
                    c=pal[2] if palette is None else pal[t],
                    alpha=alpha,
                    marker=cls.MARKERS[i],
                )
                markers_handles[n] = cls.MARKERS[i]

        if plot_reg:
            sns.regplot(
                x_gene,
                y_gene,
                data=plot_df,
                line_kws=dict(lw=1.0, color=cls.PAL_DTRACE[1]),
                marker="",
                lowess=lowess,
                truncate=True,
                ax=ax,
            )

        ax.grid(axis="both", lw=0.1, color="#e1e1e1", zorder=0)

        cor, pval = spearmanr(plot_df[x_gene], plot_df[y_gene])
        rmse = sqrt(mean_squared_error(plot_df[x_gene], plot_df[y_gene]))
        annot_text = f"Spearman's R={cor:.2g}, p-value={pval:.1e}; RMSE={rmse:.2f}"
        ax.text(0.95, 0.05, annot_text, fontsize=4, transform=ax.transAxes, ha="right")

        if plot_hue_legend and (palette is not None):
            hue_handles = [
                Line2D(
                    [0],
                    [0],
                    marker="o",
                    label=t,
                    mew=0,
                    markersize=3,
                    markerfacecolor=c,
                    lw=0,
                )
                for t, c in pal.items()
            ]
            hue_legend = ax.legend(
                handles=hue_handles,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
                prop={"size": 3},
                frameon=False,
            )
            ax.add_artist(hue_legend)

        if plot_style_legend and (style is not None):
            style_handles = [
                Line2D(
                    [0],
                    [0],
                    marker=m,
                    label=n,
                    mew=0,
                    markersize=3,
                    markerfacecolor=cls.PAL_DTRACE[2],
                    lw=0,
                )
                for n, m in markers_handles.items()
            ]
            ax.legend(
                handles=style_handles, loc="upper left", frameon=False, prop={"size": 3}
            )

        return ax

    @classmethod
    def gi_regression(
        cls,
        x_gene,
        y_gene,
        plot_df=None,
        size=None,
        size_range=None,
        size_inverse=False,
        size_legend_loc="best",
        size_legend_title=None,
        hue=None,
        style=None,
        lowess=False,
        palette=None,
        plot_reg=True,
        plot_annot=True,
        hexbin=False,
        color=None,
        label=None,
        a=0.75,
    ):
        pal = cls.PAL_DTRACE if palette is None else palette

        if plot_df is None:
            plot_df = pd.concat([x_gene, y_gene], axis=1)
            x_gene, y_gene = x_gene.name, y_gene.name

        plot_df = plot_df.dropna(subset=[x_gene, y_gene])

        if size is not None:
            plot_df = plot_df.dropna(subset=[size])

            feature_range = [1, 10] if size_range is None else size_range
            s_transform = MinMaxScaler(feature_range=feature_range)
            s_transform = s_transform.fit((plot_df[[size]] * -1) if size_inverse else plot_df[[size]])

        grid = sns.JointGrid(x_gene, y_gene, data=plot_df, space=0)

        # Joint
        if plot_reg:
            grid.plot_joint(
                sns.regplot,
                data=plot_df,
                line_kws=dict(lw=1.0, color=cls.PAL_DTRACE[1]),
                marker="",
                lowess=lowess,
                truncate=True,
            )

        hue_df = plot_df.groupby(hue) if hue is not None else [(None, plot_df)]
        for i, (h, h_df) in enumerate(hue_df):
            style_df = h_df.groupby(style) if style is not None else [(None, h_df)]

            for j, (s, s_df) in enumerate(style_df):
                if hexbin:
                    grid.ax_joint.hexbin(
                        s_df[x_gene],
                        s_df[y_gene],
                        cmap="Spectral_r",
                        gridsize=100,
                        mincnt=1,
                        bins="log",
                        lw=0,
                        alpha=1,
                    )

                else:
                    if size is None:
                        s = 3
                    elif size_inverse:
                        s = s_transform.transform(s_df[[size]] * -1)
                    else:
                        s = s_transform.transform(s_df[[size]])

                    sc = grid.ax_joint.scatter(
                        x=s_df[x_gene],
                        y=s_df[y_gene],
                        edgecolor="w",
                        lw=0.1,
                        s=s,
                        c=pal[2] if h is None else pal[h],
                        alpha=a,
                        marker=cls.MARKERS[0] if s is None else cls.MARKERS[j],
                        label=s,
                    )

                grid.x = s_df[x_gene].rename("")
                grid.y = s_df[y_gene].rename("")
                grid.plot_marginals(
                    sns.kdeplot,
                    # hist_kws=dict(linewidth=0, alpha=a),
                    cut=0,
                    legend=False,
                    shade=True,
                    color=pal[2] if h is None else pal[h],
                    label=h,
                )

        grid.ax_joint.grid(axis="both", lw=0.1, color="#e1e1e1", zorder=0)

        if plot_annot:
            cor, pval = spearmanr(plot_df[x_gene], plot_df[y_gene])
            rmse = sqrt(mean_squared_error(plot_df[x_gene], plot_df[y_gene]))
            annot_text = f"Spearman's R={cor:.2g}, p-value={pval:.1e}; RMSE={rmse:.2f}"
            grid.ax_joint.text(
                0.95,
                0.05,
                annot_text,
                fontsize=4,
                transform=grid.ax_joint.transAxes,
                ha="right",
            )

        if style is not None:
            grid.ax_joint.legend(prop=dict(size=4), frameon=False, loc=2)

        if hue is not None:
            grid.ax_marg_y.legend(
                prop=dict(size=4),
                frameon=False,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
            )

        if size is not None:
            def inverse_transform_func(x):
                arg_x = np.array(x).reshape(-1, 1)
                res = s_transform.inverse_transform(arg_x)[:, 0]
                return res

            handles, labels = sc.legend_elements(
                prop="sizes",
                num=8,
                func=inverse_transform_func,
            )
            grid.ax_joint.legend(
                handles,
                labels,
                title=size_legend_title,
                loc=size_legend_loc,
                frameon=False,
                prop={"size": 3},
            ).get_title().set_fontsize("3")

        plt.gcf().set_size_inches(2, 2)

        return grid

    @staticmethod
    def _marginal_boxplot(_, xs=None, ys=None, zs=None, vertical=False, **kws):
        if vertical:
            ax = sns.boxplot(x=zs, y=ys, orient="v", **kws)
        else:
            ax = sns.boxplot(x=xs, y=zs, orient="h", **kws)

        ax.set_ylabel("")
        ax.set_xlabel("")

    @classmethod
    def gi_regression_marginal(
        cls,
        x,
        y,
        z,
        plot_df,
        style=None,
        scatter_kws=None,
        line_kws=None,
        legend_title=None,
        discrete_pal=None,
        hue_order=None,
        annot_text=None,
        add_hline=False,
        add_vline=False,
        plot_reg=True,
        plot_annot=True,
        marginal_notch=False,
    ):
        # Defaults
        if scatter_kws is None:
            scatter_kws = dict(edgecolor="w", lw=0.3, s=8)

        if line_kws is None:
            line_kws = dict(lw=1.0, color=cls.PAL_DBGD[0])

        if discrete_pal is None:
            discrete_pal = cls.PAL_DTRACE

        if hue_order is None:
            hue_order = natsorted(set(plot_df[z]))

        #
        grid = sns.JointGrid(x, y, plot_df, space=0, ratio=8)

        grid.plot_marginals(
            cls._marginal_boxplot,
            palette=discrete_pal,
            data=plot_df,
            linewidth=0.3,
            fliersize=1,
            notch=marginal_notch,
            saturation=1.0,
            xs=x,
            ys=y,
            zs=z,
            showcaps=False,
            boxprops=cls.BOXPROPS,
            whiskerprops=cls.WHISKERPROPS,
            flierprops=cls.FLIERPROPS,
            medianprops=dict(linestyle="-", linewidth=1.0),
        )

        if plot_reg:
            sns.regplot(
                x=x,
                y=y,
                data=plot_df,
                color=discrete_pal[0],
                truncate=True,
                fit_reg=True,
                scatter=False,
                line_kws=line_kws,
                ax=grid.ax_joint,
            )

        for feature in hue_order:
            dfs = plot_df[plot_df[z] == feature]
            dfs = (
                dfs.assign(style=1).groupby("style")
                if style is None
                else dfs.groupby(style)
            )

            for i, (mtype, df) in enumerate(dfs):
                sns.regplot(
                    x=x,
                    y=y,
                    data=df,
                    color=discrete_pal[feature],
                    fit_reg=False,
                    scatter_kws=scatter_kws,
                    label=mtype if i > 0 else None,
                    marker=cls.MARKERS[i],
                    ax=grid.ax_joint,
                )

        if style is not None:
            grid.ax_joint.legend(prop=dict(size=4), frameon=False, loc=2)

        # Annotation
        if plot_annot:
            if annot_text is None:
                df_corr = plot_df.dropna(subset=[x, y, z])
                cor, pval = pearsonr(df_corr[x], df_corr[y])
                annot_text = f"R={cor:.2g}, p={pval:.1e}"

            grid.ax_joint.text(
                0.95,
                0.05,
                annot_text,
                fontsize=4,
                transform=grid.ax_joint.transAxes,
                ha="right",
            )

        if add_hline:
            grid.ax_joint.axhline(0, ls="-", lw=0.3, c=cls.PAL_DBGD[0], alpha=0.2)

        if add_vline:
            grid.ax_joint.axvline(0, ls="-", lw=0.3, c=cls.PAL_DBGD[0], alpha=0.2)

        handles = [
            mpatches.Circle(
                (0.0, 0.0),
                0.25,
                facecolor=discrete_pal[t],
                label=f"{t} (N={(plot_df[z] == t).sum()})",
            )
            for t in hue_order
        ]

        grid.ax_marg_y.legend(
            handles=handles,
            title=z if legend_title is None else legend_title,
            loc="center left",
            bbox_to_anchor=(1, 0.5),
            frameon=False,
        )

        grid.ax_joint.grid(axis="both", lw=0.1, color="#e1e1e1", zorder=0)

        plt.gcf().set_size_inches(1.5, 1.5)

        return grid

    @classmethod
    def gi_classification(
        cls,
        x_gene,
        y_gene,
        plot_df,
        hue=None,
        palette=None,
        orient="v",
        stripplot=True,
        notch=True,
        order=None,
        hue_order=None,
        plot_legend=True,
        legend_kws=None,
        ax=None,
    ):
        pal = cls.PAL_DTRACE if palette is None else palette

        if ax is None and orient == "v":
            figsize = (0.2 * len(set(plot_df[x_gene])), 2)

        elif ax is None and orient == "h":
            figsize = (2, 0.2 * len(set(plot_df[y_gene])))

        else:
            figsize = None

        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=figsize, dpi=600)

        if stripplot:
            sns.stripplot(
                x=x_gene,
                y=y_gene,
                order=order,
                hue=hue,
                hue_order=hue_order,
                data=plot_df,
                dodge=True,
                orient=orient,
                jitter=0.3,
                size=1.5,
                linewidth=0.1,
                alpha=0.5,
                edgecolor="white",
                palette=pal,
                ax=ax,
                zorder=0,
            )

        bp = sns.boxplot(
            x=x_gene,
            y=y_gene,
            order=order,
            hue=hue,
            hue_order=hue_order,
            data=plot_df,
            orient=orient,
            notch=notch,
            boxprops=dict(linewidth=0.3),
            whiskerprops=dict(linewidth=0.3),
            medianprops=cls.MEDIANPROPS,
            flierprops=cls.FLIERPROPS,
            palette=pal,
            showcaps=False,
            sym="" if stripplot else None,
            saturation=1.0,
            ax=ax,
        )

        for patch in bp.artists:
            r, g, b, a = patch.get_facecolor()
            patch.set_facecolor((r, g, b, 0.1))

        ax.grid(
            True,
            axis="y" if orient == "v" else "x",
            ls="-",
            lw=0.1,
            alpha=1.0,
            zorder=0,
        )

        if plot_legend and (hue is not None):
            hue_nfeatures = len(set(plot_df[hue]))
            handles, labels = bp.get_legend_handles_labels()
            ax.legend(
                handles[: (hue_nfeatures - 1)],
                labels[: (hue_nfeatures - 1)],
                frameon=False,
                **legend_kws,
            )

        elif ax.get_legend() is not None:
            ax.get_legend().remove()

        return ax

    @classmethod
    def gi_tissue_plot(
        cls,
        x,
        y,
        plot_df=None,
        hue="tissue",
        pal=CrispyPlot.PAL_TISSUE_2,
        plot_reg=True,
        annot=True,
        lowess=False,
        figsize=(2, 2),
    ):
        if plot_df is None:
            plot_df = pd.concat([x, y], axis=1)
            x, y = x.name, y.name

        plot_df = plot_df.dropna(subset=[x, y, hue])

        fig, ax = plt.subplots(figsize=figsize, dpi=600)

        for t, df in plot_df.groupby(hue):
            ax.scatter(
                df[x].values,
                df[y].values,
                c=pal[t],
                marker="o",
                linewidths=0,
                s=5,
                label=t,
                alpha=0.8,
            )

        if plot_reg:
            sns.regplot(
                x,
                y,
                data=plot_df,
                line_kws=dict(lw=1.0, color=cls.PAL_DTRACE[1]),
                marker="",
                lowess=lowess,
                truncate=True,
                ax=ax,
            )

        if annot:
            cor, pval = spearmanr(plot_df[x], plot_df[y])
            annot_text = f"Spearman's R={cor:.2g}, p-value={pval:.1e}"
            ax.text(
                0.95, 0.05, annot_text, fontsize=4, transform=ax.transAxes, ha="right"
            )

        ax.set_xlabel(x)
        ax.set_ylabel(y)
        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)
        ax.legend(
            loc="center left",
            bbox_to_anchor=(1, 0.5),
            prop={"size": 3},
            frameon=False,
            title=hue,
        ).get_title().set_fontsize("3")

        return ax

    @classmethod
    def gi_continuous_plot(
        cls,
        x,
        y,
        z,
        plot_df,
        cmap="Spectral_r",
        joint_alpha=0.8,
        mid_point_norm=True,
        mid_point=0,
        cbar_label=None,
        lowess=False,
        plot_reg=False,
        corr_annotation=True,
        ax=None,
    ):
        df = plot_df.dropna(subset=[x, y, z])

        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(2.5, 2), dpi=600)

        sc = ax.scatter(
            df[x],
            df[y],
            c=df[z],
            marker="o",
            edgecolor="",
            s=3,
            cmap=cmap,
            alpha=joint_alpha,
            norm=MidpointNormalize(midpoint=mid_point) if mid_point_norm else None,
        )

        cbar = plt.colorbar(sc)
        cbar.ax.set_ylabel(
            z if cbar_label is None else cbar_label, rotation=270, va="bottom"
        )

        if plot_reg:
            sns.regplot(
                x,
                y,
                data=plot_df,
                line_kws=dict(lw=1.0, color=cls.PAL_DTRACE[1]),
                marker="",
                lowess=lowess,
                truncate=True,
                ax=ax,
            )

        if corr_annotation:
            cor, pval = spearmanr(df[x], df[y])
            annot_text = f"R={cor:.2g}, p={pval:.1e}"
            ax.text(
                0.95, 0.05, annot_text, fontsize=4, transform=ax.transAxes, ha="right"
            )

        ax.set_xlabel(x)
        ax.set_ylabel(y)
        ax.grid(True, ls="-", lw=0.1, alpha=1.0, zorder=0)

        return ax

    @classmethod
    def gi_manhattan(cls, assoc_df):
        _, axs = plt.subplots(1, len(Utils.CHR_ORDER), figsize=(len(Utils.CHR_ORDER), 3), dpi=600, sharey="row")

        for i, ctype in enumerate(Utils.CHR_ORDER):
            ax = axs[i]

            for ttype in ["ppi != 'T'", "ppi == 'T'"]:
                t_df = assoc_df.query(f"(chr == '{ctype}') & (fdr < .1) & ({ttype})")
                t_df["log_pval"] = -np.log10(t_df["pval"])

                def calculate_score(pval, beta):
                    s = np.log10(pval)
                    if beta > 0:
                        s *= -1
                    return s

                t_df["score"] = [
                    calculate_score(p, b) for p, b in t_df[["pval", "beta"]].values
                ]

                ax.scatter(
                    t_df["chr_pos"],
                    t_df["score"],
                    s=5,
                    alpha=.7,
                    c=CrispyPlot.PAL_DTRACE[0],
                    lw=.5 if ttype == "ppi == 'T'" else 0,
                    edgecolors=CrispyPlot.PAL_DTRACE[1] if ttype == "ppi == 'T'" else None,
                )

                labels = [
                    ax.text(
                        row["chr_pos"],
                        row["score"],
                        f"{row['y_id'].split(';')[1] if ';' in row['y_id'] else row['y_id']} ~ {row['x_id']}",
                        color="k",
                        fontsize=4,
                        zorder=3,
                    )
                    for _, row in t_df.query(f"log_pval > 10").head(5).iterrows()
                ]
                adjust_text(
                    labels,
                    arrowprops=dict(arrowstyle="-", color="k", alpha=0.75, lw=0.3),
                    ax=ax,
                )

            ax.set_xlabel(f"Chr {ctype}")
            ax.set_ylabel(f"-log10(p-value)" if i == 0 else None)
            ax.grid(axis="y", lw=0.1, color="#e1e1e1", zorder=0)

        plt.subplots_adjust(hspace=0, wspace=0.05)